# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import os
import pytest

from numpy.testing import (
    assert_almost_equal,
    assert_equal,
)

from MDAnalysisTests.coordinates.reference import RefAdKSmall
from MDAnalysisTests.coordinates.base import _SingleFrameReader
from MDAnalysisTests.datafiles import PQR
from MDAnalysisTests import make_Universe


class TestPQRReader(_SingleFrameReader):
    __test__ = True
    def setUp(self):
        self.universe = mda.Universe(PQR)
        # 3 decimals in PDB spec
        # http://www.wwpdb.org/documentation/format32/sect9.html#ATOM
        self.prec = 3

    def test_total_charge(self):
        assert_almost_equal(
            self.universe.atoms.total_charge(), self.ref_charmm_totalcharge, 3,
            "Total charge (in CHARMM) does not match expected value.")

    def test_hydrogenCharges(self):
        assert_almost_equal(self.universe.select_atoms('name H').charges,
                            self.ref_charmm_Hcharges, 3,
                            "Charges for H atoms do not match.")

    # Note that the whole system gets the sysID 'SYSTEM' for the PQR file (when
    # read with a PSF it is 's4AKE')
    def test_ArgCACharges(self):
        ag = self.universe.select_atoms('resname ARG and name CA')
        assert_almost_equal(
            ag.charges, self.ref_charmm_ArgCAcharges,
            3, "Charges for CA atoms in Arg residues do not match.")

    def test_ProNCharges(self):
        ag = self.universe.select_atoms('resname PRO and name N')
        assert_almost_equal(
            ag.charges, self.ref_charmm_ProNcharges, 3,
            "Charges for N atoms in Pro residues do not match.")

    def test_dimensions(self):
        # Issue #3327 - dimensions should always be set to None
        assert self.universe.dimensions is None


class TestPQRWriter(RefAdKSmall):
    @staticmethod
    @pytest.fixture
    def universe():
        return mda.Universe(PQR)

    prec = 3

    def test_writer_noChainID(self, universe, tmpdir):
        outfile = str(tmpdir.join('pqr-test.pqr'))

        assert_equal(universe.segments.segids[0], 'SYSTEM')
        universe.atoms.write(outfile)
        u = mda.Universe(outfile)
        assert_equal(u.segments.segids[0], 'SYSTEM')
        assert_almost_equal(u.atoms.positions,
                            universe.atoms.positions, self.prec,
                            err_msg="Writing PQR file with PQRWriter does "
                            "not reproduce original coordinates")
        assert_almost_equal(u.atoms.charges, universe.atoms.charges,
                            self.prec, err_msg="Writing PQR file with "
                            "PQRWriter does not reproduce original charges")
        assert_almost_equal(u.atoms.radii, universe.atoms.radii,
                            self.prec, err_msg="Writing PQR file with "
                            "PQRWriter does not reproduce original radii")

    # 363 TODO:
    # Not sure if this should be a segid or chainID?
    # Topology system now allows for both of these
    def test_write_withChainID(self, universe, tmpdir):
        outfile = str(tmpdir.join('pqr-test.pqr'))

        universe.segments.segids = 'A'
        assert_equal(universe.segments.segids[0], 'A')  # sanity check
        universe.atoms.write(outfile)
        u = mda.Universe(outfile)
        assert_equal(u.segments.segids[0], 'A')
        assert_almost_equal(u.atoms.positions,
                            universe.atoms.positions, self.prec,
                            err_msg="Writing PQR file with PQRWriter does "
                            "not reproduce original coordinates")
        assert_almost_equal(u.atoms.charges, universe.atoms.charges,
                            self.prec, err_msg="Writing PQR file with "
                            "PQRWriter does not reproduce original charges")
        assert_almost_equal(u.atoms.radii, universe.atoms.radii,
                            self.prec, err_msg="Writing PQR file with "
                            "PQRWriter does not reproduce original radii")

    def test_timestep_not_modified_by_writer(self, universe, tmpdir):
        outfile = str(tmpdir.join('pqr-test.pqr'))

        ts = universe.trajectory.ts
        x = ts.positions.copy()
        universe.atoms.write(outfile)
        assert_equal(ts.positions, x,
                     err_msg="Positions in Timestep were modified by writer.")

    def test_total_charge(self, universe, tmpdir):
        outfile = str(tmpdir.join('pqr-test.pqr'))
        universe.atoms.write(outfile)
        u = mda.Universe(outfile)
        assert_almost_equal(
            u.atoms.total_charge(), self.ref_charmm_totalcharge, 3,
            "Total charge (in CHARMM) does not match expected value.")

class TestPQRWriterMissingAttrs(object):
    # pqr requires names, resids, resnames, segids, radii, charges
    @staticmethod
    @pytest.fixture
    def reqd_attributes():
        return ['names', 'resids', 'resnames', 'radii', 'charges']

    @staticmethod
    @pytest.fixture
    def outfile(tmpdir):
        return str(tmpdir.join('pqr-writer-test.pqr'))

    def test_no_names_writing(self, reqd_attributes, outfile):
        attrs = reqd_attributes
        attrs.remove('names')
        u = make_Universe(attrs, trajectory=True)

        with pytest.warns(UserWarning):
            u.atoms.write(outfile)

        u2 = mda.Universe(outfile)

        assert all(u2.atoms.names == 'X')

    def test_no_resnames_writing(self, reqd_attributes, outfile):
        attrs = reqd_attributes
        attrs.remove('resnames')
        u = make_Universe(attrs, trajectory=True)

        with pytest.warns(UserWarning):
            u.atoms.write(outfile)

        u2 = mda.Universe(outfile)

        assert all(u2.residues.resnames == 'UNK')

    def test_no_radii_writing(self, reqd_attributes, outfile):
        attrs = reqd_attributes
        attrs.remove('radii')
        u = make_Universe(attrs, trajectory=True)

        with pytest.warns(UserWarning):
            u.atoms.write(outfile)

        u2 = mda.Universe(outfile)

        assert all(u2.atoms.radii == 1.0)

    def test_no_charges_writing(self, reqd_attributes, outfile):
        attrs = reqd_attributes
        attrs.remove('charges')
        u = make_Universe(attrs, trajectory=True)

        with pytest.warns(UserWarning):
            u.atoms.write(outfile)

        u2 = mda.Universe(outfile)

        assert all(u2.atoms.charges == 0.0)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import MDAnalysis as mda
from MDAnalysisTests.util import import_not_available


requires_rdkit = pytest.mark.skipif(import_not_available("rdkit"),
                                    reason="requires RDKit")


@requires_rdkit
class TestConvertTo:
    @pytest.fixture(scope="class")
    def u(self):
        return mda.Universe.from_smiles("CCO")

    def test_convert_to_case_insensitive(self, u):
        mol = u.atoms.convert_to("rdkit")

    def test_convert_to_lib_as_method(self, u):
        mol = u.atoms.convert_to.rdkit()

    def test_convert_to_kwargs(self, u):
        mol = u.atoms.convert_to("RDKIT", NoImplicit=False)
        assert mol.GetAtomWithIdx(0).GetNoImplicit() is False

    def test_convert_to_lib_method_kwargs(self, u):
        mol = u.atoms.convert_to.rdkit(NoImplicit=False)
        assert mol.GetAtomWithIdx(0).GetNoImplicit() is False


class TestAccessor:
    def test_access_from_class(self):
        assert (mda.core.AtomGroup.convert_to is
                mda.core.accessors.ConverterWrapper)


class TestConverterWrapper:
    def test_raises_valueerror(self):
        u = mda.Universe.empty(1)
        with pytest.raises(ValueError,
                           match="No 'mdanalysis' converter found"):
            u.atoms.convert_to("mdanalysis")

    @requires_rdkit
    def test_single_instance(self):
        u1 = mda.Universe.from_smiles("C")
        u2 = mda.Universe.from_smiles("CC")
        assert (u1.atoms.convert_to.rdkit.__wrapped__ is
                u2.atoms.convert_to.rdkit.__wrapped__)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import itertools
import pickle

import numpy as np
import pytest
from unittest import TestCase
from numpy.testing import (assert_equal, assert_almost_equal,
                           assert_array_almost_equal, assert_allclose)

import MDAnalysis as mda
from MDAnalysis.coordinates.base import Timestep
from MDAnalysis.transformations import translate
from MDAnalysis import NoDataError


from MDAnalysisTests.coordinates.reference import RefAdKSmall
from MDAnalysisTests.datafiles import AUX_XVG_HIGHF, AUX_XVG_LOWF
from MDAnalysisTests import make_Universe


class _SingleFrameReader(TestCase, RefAdKSmall):
    # see TestPDBReader how to set up!
    __test__ = False

    def tearDown(self):
        del self.universe

    def test_load_file(self):
        U = self.universe
        assert_equal(len(U.atoms), self.ref_n_atoms,
                     "load Universe from file {0!s}".format(U.trajectory.filename))
        assert_equal(U.atoms.select_atoms('resid 150 and name HA2').atoms[0],
                     U.atoms[self.ref_E151HA2_index], "Atom selections")

    def test_n_atoms(self):
        assert_equal(self.universe.trajectory.n_atoms, self.ref_n_atoms,
                     "wrong number of atoms")

    def test_numres(self):
        assert_equal(self.universe.atoms.n_residues, 214,
                     "wrong number of residues")

    def test_n_frames(self):
        assert_equal(self.universe.trajectory.n_frames, 1,
                     "wrong number of frames in pdb")

    def test_time(self):
        assert_equal(self.universe.trajectory.time, 0.0,
                     "wrong time of the frame")

    def test_frame(self):
        assert_equal(
            self.universe.trajectory.frame, 0,
            "wrong frame number (0-based, should be 0 for single frame "
            "readers)")

    def test_frame_index_0(self):
        self.universe.trajectory[0]
        assert_equal(self.universe.trajectory.ts.frame, 0,
                     "frame number for frame index 0 should be 0")

    def test_frame_index_1_raises_IndexError(self):
        def go_to_2(traj=self.universe.trajectory):
            traj[1]
        with pytest.raises(IndexError):
            go_to_2()

    def test_dt(self):
        """testing that accessing universe.trajectory.dt gives 1.0
        (the default)"""
        assert_equal(1.0, self.universe.trajectory.dt)

    def test_coordinates(self):
        A10CA = self.universe.select_atoms('name CA')[10]
        # restrict accuracy to maximum in PDB files (3 decimals)
        assert_almost_equal(A10CA.position,
                            self.ref_coordinates['A10CA'],
                            3,
                            err_msg="wrong coordinates for A10:CA")

    def test_distances(self):
        NTERM = self.universe.select_atoms('name N')[0]
        CTERM = self.universe.select_atoms('name C')[-1]
        d = mda.lib.mdamath.norm(NTERM.position - CTERM.position)
        assert_almost_equal(d,
                            self.ref_distances['endtoend'],
                            self.prec,
                            err_msg="distance between M1:N and G214:C")

    def test_full_slice(self):
        trj_iter = self.universe.trajectory[:]
        frames = [ts.frame for ts in trj_iter]
        assert_equal(frames, np.arange(self.universe.trajectory.n_frames))

    def test_last_slice(self):
        # should be same as above: only 1 frame!
        trj_iter = self.universe.trajectory[-1:]
        frames = [ts.frame for ts in trj_iter]
        assert_equal(frames, np.arange(self.universe.trajectory.n_frames))

    def test_pickle_singleframe_reader(self):
        reader = self.universe.trajectory
        reader_p = pickle.loads(pickle.dumps(reader))
        assert_equal(len(reader), len(reader_p))
        assert_equal(reader.ts, reader_p.ts,
                     "Single-frame timestep is changed after pickling")


class BaseReference(object):
    def __init__(self):
        self.trajectory = None
        self.n_atoms = 5
        self.n_frames = 5
        # default for the numpy test functions
        self.prec = 6
        self.container_format = False
        self.changing_dimensions = False

        # for testing auxiliary addition
        self.aux_lowf = AUX_XVG_LOWF  # test auxiliary with lower frequency
        self.aux_lowf_dt = 2  # has steps at 0ps, 2ps, 4ps
        # representative data for each trajectory frame, assuming 'closest' option
        self.aux_lowf_data = [[2 ** 0],  # frame 0 = 0ps = step 0
                              [np.nan],  # frame 1 = 1ps = no step
                              [2 ** 1],  # frame 2 = 2ps = step 1
                              [np.nan],  # frame 3 = 3ps = no step
                              [2 ** 2],  # frame 4 = 4ps = step 2
                              ]
        self.aux_lowf_frames_with_steps = [0, 2, 4]  # trajectory frames with
        # corresponding auxiliary steps

        self.aux_highf = AUX_XVG_HIGHF  # test auxiliary with higher frequency
        self.aux_highf_dt = 0.5  # has steps at 0, 0.5, 1, ... 3.5, 4ps
        self.aux_highf_data = [[2 ** 0],  # frame 0 = 0ps = step 0
                               [2 ** 2],  # frame 1 = 1ps = step 2
                               [2 ** 4],  # frame 2 = 2ps = step 4
                               [2 ** 6],  # frame 3 = 3ps = step 6
                               [2 ** 8],  # frame 4 = 4ps = step 8
                               ]
        self.aux_highf_n_steps = 10
        self.aux_highf_all_data = [[2 ** i] for i in range(self.aux_highf_n_steps)]

        self.aux_offset_by = 0.25

        self.first_frame = Timestep(self.n_atoms)
        self.first_frame.positions = np.arange(
            3 * self.n_atoms).reshape(self.n_atoms, 3)
        self.first_frame.frame = 0
        self.first_frame.aux.lowf = self.aux_lowf_data[0]
        self.first_frame.aux.highf = self.aux_highf_data[0]

        self.second_frame = self.first_frame.copy()
        self.second_frame.positions = 2 ** 1 * self.first_frame.positions
        self.second_frame.frame = 1
        self.second_frame.aux.lowf = self.aux_lowf_data[1]
        self.second_frame.aux.highf = self.aux_highf_data[1]

        self.last_frame = self.first_frame.copy()
        self.last_frame.positions = 2 ** 4 * self.first_frame.positions
        self.last_frame.frame = self.n_frames - 1
        self.last_frame.aux.lowf = self.aux_lowf_data[-1]
        self.last_frame.aux.highf = self.aux_highf_data[-1]

        # remember frames are 0 indexed
        self.jump_to_frame = self.first_frame.copy()
        self.jump_to_frame.positions = 2 ** 3 * self.first_frame.positions
        self.jump_to_frame.frame = 3
        self.jump_to_frame.aux.lowf = self.aux_lowf_data[3]
        self.jump_to_frame.aux.highf = self.aux_highf_data[3]

        self.dimensions = np.array([81.1, 82.2, 83.3, 75, 80, 85],
                                   dtype=np.float32)
        self.dimensions_second_frame = np.array([82.1, 83.2, 84.3, 75.1, 80.1,
                                                 85.1], dtype=np.float32)
        self.volume = mda.lib.mdamath.box_volume(self.dimensions)
        self.time = 0
        self.dt = 1
        self.totaltime = 4

    def iter_ts(self, i):
        ts = self.first_frame.copy()
        ts.positions = 2 ** i * self.first_frame.positions
        ts.time = i
        ts.frame = i
        ts.aux.lowf = np.array(self.aux_lowf_data[i])
        ts.aux.highf = np.array(self.aux_highf_data[i])
        return ts


class BaseReaderTest(object):
    @staticmethod
    @pytest.fixture()
    def reader(ref):
        reader = ref.reader(ref.trajectory)
        reader.add_auxiliary('lowf', ref.aux_lowf, dt=ref.aux_lowf_dt, initial_time=0, time_selector=None)
        reader.add_auxiliary('highf', ref.aux_highf, dt=ref.aux_highf_dt, initial_time=0, time_selector=None)
        return reader

    @staticmethod
    @pytest.fixture()
    def transformed(ref):
        transformed = ref.reader(ref.trajectory)
        transformed.add_transformations(translate([1,1,1]), translate([0,0,0.33]))
        return transformed

    def test_n_atoms(self, ref, reader):
        assert_equal(reader.n_atoms, ref.n_atoms)

    def test_n_frames(self, ref, reader):
        assert_equal(len(reader), ref.n_frames)

    def test_first_frame(self, ref, reader):
        reader.rewind()
        assert_timestep_almost_equal(reader.ts, ref.first_frame,
                                     decimal=ref.prec)

    def test_double_close(self, reader):
        reader.close()
        reader.close()
        reader._reopen()

    def test_get_writer_1(self, ref, reader, tmpdir):
        with tmpdir.as_cwd():
            outfile = 'test-writer.' + ref.ext
            with reader.Writer(outfile) as W:
                assert_equal(isinstance(W, ref.writer), True)
                assert_equal(W.n_atoms, reader.n_atoms)

    def test_get_writer_2(self, ref, reader, tmpdir):
        with tmpdir.as_cwd():
            outfile = 'test-writer.' + ref.ext
            with reader.Writer(outfile, n_atoms=100) as W:
                assert_equal(isinstance(W, ref.writer), True)
                assert_equal(W.n_atoms, 100)

    def test_dt(self, ref, reader):
        assert_almost_equal(reader.dt, ref.dt, decimal=ref.prec)

    def test_ts_dt_matches_reader(self, reader):
        assert_equal(reader.ts.dt, reader.dt)

    def test_total_time(self, ref, reader):
        assert_almost_equal(reader.totaltime, ref.totaltime, decimal=ref.prec)

    def test_first_dimensions(self, ref, reader):
        reader.rewind()
        if ref.dimensions is None:
            assert reader.ts.dimensions is None
        else:
            assert_array_almost_equal(reader.ts.dimensions,
                                      ref.dimensions,
                                      decimal=ref.prec)

    def test_changing_dimensions(self, ref, reader):
        if ref.changing_dimensions:
            reader.rewind()
            if ref.dimensions is None:
                assert reader.ts.dimensions is None
            else:
                assert_array_almost_equal(reader.ts.dimensions,
                                          ref.dimensions,
                                          decimal=ref.prec)
            reader[1]
            if ref.dimensions_second_frame is None:
                assert reader.ts.dimensions is None
            else:
                assert_array_almost_equal(reader.ts.dimensions,
                                          ref.dimensions_second_frame,
                                          decimal=ref.prec)

    def test_volume(self, ref, reader):
        reader.rewind()
        vol = reader.ts.volume
        # Here we can only be sure about the numbers upto the decimal point due
        # to floating point impressions.
        assert_almost_equal(vol, ref.volume, 0)

    def test_iter(self, ref, reader):
        for i, ts in enumerate(reader):
            assert_timestep_almost_equal(ts, ref.iter_ts(i),
                                         decimal=ref.prec)

    def test_add_same_auxname_raises_ValueError(self, ref, reader):
        with pytest.raises(ValueError):
            reader.add_auxiliary('lowf', ref.aux_lowf)

    def test_remove_auxiliary(self, reader):
        reader.remove_auxiliary('lowf')
        with pytest.raises(AttributeError):
            getattr(reader._auxs, 'lowf')
        with pytest.raises(AttributeError):
            getattr(reader.ts.aux, 'lowf')

    def test_remove_nonexistant_auxiliary_raises_ValueError(self, reader):
        with pytest.raises(ValueError):
            reader.remove_auxiliary('nonexistant')

    def test_iter_auxiliary(self, ref, reader):
        # should go through all steps in 'highf'
        for i, auxstep in enumerate(reader.iter_auxiliary('highf')):
            assert_almost_equal(auxstep.data, ref.aux_highf_all_data[i],
                                err_msg="Auxiliary data does not match for "
                                        "step {}".format(i))

    def test_get_aux_attribute(self, ref, reader):
        assert_equal(reader.get_aux_attribute('lowf', 'dt'),
                     ref.aux_lowf_dt)

    def test_iter_as_aux_cutoff(self, ref, reader):
        # load an auxiliary with the same dt but offset from trajectory, and a
        # cutoff of 0
        reader.add_auxiliary('offset', ref.aux_lowf,
                                  dt=ref.dt, time_selector=None,
                                  initial_time=ref.aux_offset_by,
                                  cutoff=0)
        # no auxiliary steps will fall within the cutoff for any frame, so
        # iterating using iter_as_aux should give us nothing
        num_frames = len([i for i in reader.iter_as_aux('offset')])
        assert_equal(num_frames, 0, "iter_as_aux should iterate over 0 frames,"
                                    " not {}".format(num_frames))

    def test_reload_auxiliaries_from_description(self, ref, reader):
        # get auxiliary desscriptions form existing reader
        descriptions = reader.get_aux_descriptions()
        # load a new reader, without auxiliaries
        reader = ref.reader(ref.trajectory)
        # load auxiliaries into new reader, using description...
        for aux in descriptions:
            reader.add_auxiliary(**aux)
        # should have the same number of auxiliaries
        assert_equal(reader.aux_list, reader.aux_list,
                     'Number of auxiliaries does not match')
        # each auxiliary should be the same
        for auxname in reader.aux_list:
            assert_equal(reader._auxs[auxname], reader._auxs[auxname],
                         'AuxReaders do not match')

    def test_stop_iter(self, reader):
        # reset to 0
        reader.rewind()
        for ts in reader[:-1]:
            pass
        assert_equal(reader.frame, 0)

    def test_transformations_iter(self, ref, transformed):
        # test one iteration and see if the transformations are applied
        v1 = np.float32((1,1,1))
        v2 = np.float32((0,0,0.33))
        for i, ts in enumerate(transformed):
            idealcoords = ref.iter_ts(i).positions + v1 + v2
            assert_array_almost_equal(ts.positions, idealcoords, decimal=ref.prec)

    def test_transformations_2iter(self, ref, transformed):
        # Are the transformations applied and
        # are the coordinates "overtransformed"?
        v1 = np.float32((1,1,1))
        v2 = np.float32((0,0,0.33))
        idealcoords=[]
        for i, ts in enumerate(transformed):
            idealcoords.append(ref.iter_ts(i).positions + v1 + v2)
            assert_array_almost_equal(ts.positions, idealcoords[i], decimal=ref.prec)

        for i, ts in enumerate(transformed):
            assert_almost_equal(ts.positions, idealcoords[i], decimal=ref.prec)

    def test_transformations_slice(self, ref, transformed):
        # Are the transformations applied when iterating over a slice of the trajectory?
        v1 = np.float32((1,1,1))
        v2 = np.float32((0,0,0.33))
        for i,ts in enumerate(transformed[2:3:1]):
            idealcoords = ref.iter_ts(ts.frame).positions + v1 + v2
            assert_array_almost_equal(ts.positions, idealcoords, decimal = ref.prec)

    def test_transformations_switch_frame(self, ref, transformed):
        # This test checks if the transformations are applied and if the coordinates
        # "overtransformed" on different situations
        # Are the transformations applied when we switch to a different frame?
        v1 = np.float32((1,1,1))
        v2 = np.float32((0,0,0.33))
        first_ideal = ref.iter_ts(0).positions + v1 + v2
        if len(transformed)>1:
            assert_array_almost_equal(transformed[0].positions, first_ideal, decimal = ref.prec)
            second_ideal = ref.iter_ts(1).positions + v1 + v2
            assert_array_almost_equal(transformed[1].positions, second_ideal, decimal = ref.prec)

            # What if we comeback to the previous frame?
            assert_array_almost_equal(transformed[0].positions, first_ideal, decimal = ref.prec)

            # How about we switch the frame to itself?
            assert_array_almost_equal(transformed[0].positions, first_ideal, decimal = ref.prec)
        else:
            assert_array_almost_equal(transformed[0].positions, first_ideal, decimal = ref.prec)

    def test_transformation_rewind(self,ref, transformed):
        # this test checks if the transformations are applied after rewinding the
        # trajectory
        v1 = np.float32((1,1,1))
        v2 = np.float32((0,0,0.33))
        ideal_coords = ref.iter_ts(0).positions + v1 + v2
        transformed.rewind()
        assert_array_almost_equal(transformed[0].positions, ideal_coords, decimal = ref.prec)

    def test_transformations_copy(self,ref,transformed):
        # this test checks if transformations are carried over a copy and if the
        # coordinates of the copy are equal to the original's
        v1 = np.float32((1,1,1))
        v2 = np.float32((0,0,0.33))
        new = transformed.copy()
        assert_equal(transformed.transformations, new.transformations,
                     "transformations are not equal")
        for i, ts in enumerate(new):
            ideal_coords = ref.iter_ts(i).positions + v1 + v2
            assert_array_almost_equal(ts.positions, ideal_coords, decimal = ref.prec)

    def test_add_another_transformations_raises_ValueError(self, transformed):
        # After defining the transformations, the workflow cannot be changed
        with pytest.raises(ValueError):
            transformed.add_transformations(translate([2,2,2]))

    def test_pickle_reader(self, reader):
        reader_p = pickle.loads(pickle.dumps(reader))
        assert_equal(len(reader), len(reader_p))
        assert_equal(reader.ts, reader_p.ts,
                     "Timestep is changed after pickling")


class MultiframeReaderTest(BaseReaderTest):
    def test_last_frame(self, ref, reader):
        ts = reader[-1]
        assert_timestep_almost_equal(ts, ref.last_frame,
                                     decimal=ref.prec)

    def test_go_over_last_frame(self, ref, reader):
        with pytest.raises(IndexError):
            reader[ref.n_frames + 1]

    def test_frame_jump(self, ref, reader):
        ts = reader[ref.jump_to_frame.frame]
        assert_timestep_almost_equal(ts, ref.jump_to_frame,
                                     decimal=ref.prec)

    def test_frame_jump_issue1942(self, ref, reader):
        """Test for issue 1942 (especially XDR on macOS)"""
        reader.rewind()
        try:
            for ii in range(ref.n_frames + 2):
                reader[0]
        except StopIteration:
            pytest.fail("Frame-seeking wrongly iterated (#1942)")

    def test_next_gives_second_frame(self, ref, reader):
        reader = ref.reader(ref.trajectory)
        ts = reader.next()
        assert_timestep_almost_equal(ts, ref.second_frame,
                                     decimal=ref.prec)

    def test_reopen(self, ref, reader):
        reader.close()
        reader._reopen()
        ts = reader.next()
        assert_timestep_almost_equal(ts, ref.first_frame,
                                     decimal=ref.prec)

    def test_rename_aux(self, ref, reader):
        reader.rename_aux('lowf', 'lowf_renamed')
        # data should now be in aux namespace under new name
        assert_equal(reader.ts.aux.lowf_renamed,
                    ref.aux_lowf_data[0])
        # old name should be removed
        with pytest.raises(AttributeError):
            getattr(reader.ts.aux, 'lowf')
        # new name should be retained
        next(reader)
        assert_equal(reader.ts.aux.lowf_renamed,
                     ref.aux_lowf_data[1])

    def test_iter_as_aux_highf(self, ref, reader):
        # auxiliary has a higher frequency, so iter_as_aux should behave the
        # same as regular iteration over the trjectory
        for i, ts in enumerate(reader.iter_as_aux('highf')):
            assert_timestep_almost_equal(ts, ref.iter_ts(i),
                                         decimal=ref.prec)

    def test_iter_as_aux_lowf(self, ref, reader):
        # auxiliary has a lower frequency, so iter_as_aux should iterate over
        # only frames where there is a corresponding auxiliary value
        for i, ts in enumerate(reader.iter_as_aux('lowf')):
            assert_timestep_almost_equal(ts,
                                         ref.iter_ts(ref.aux_lowf_frames_with_steps[i]),
                                         decimal=ref.prec)

    @pytest.mark.parametrize("accessor", [
              lambda traj: traj[[0, 1, 2]],
              lambda traj: traj[:3],
              lambda traj: traj],
              ids=["indexed", "sliced", "all"])
    def test_iter_rewinds(self, reader, accessor):
        for ts_indices in accessor(reader):
            pass
        assert_equal(ts_indices.frame, 0)

    #  To make sure we not only save the current timestep information,
    #  but also maintain its relative position.
    def test_pickle_next_ts_reader(self, reader):
        reader_p = pickle.loads(pickle.dumps(reader))
        assert_equal(next(reader), next(reader_p),
                     "Next timestep is changed after pickling")

    #  To make sure pickle works for last frame.
    def test_pickle_last_ts_reader(self, reader):
        #  move current ts to last frame.
        reader[-1]
        reader_p = pickle.loads(pickle.dumps(reader))
        assert_equal(len(reader), len(reader_p),
                     "Last timestep is changed after pickling")
        assert_equal(reader.ts, reader_p.ts,
                     "Last timestep is changed after pickling")


class BaseWriterTest(object):
    @staticmethod
    @pytest.fixture()
    def reader(ref):
        return ref.reader(ref.trajectory)

    @staticmethod
    @pytest.fixture()
    def u_no_resnames():
        return make_Universe(['names', 'resids'], trajectory=True)

    @staticmethod
    @pytest.fixture()
    def u_no_resids():
        return make_Universe(['names', 'resnames'], trajectory=True)

    @staticmethod
    @pytest.fixture()
    def u_no_names():
        return make_Universe(['resids', 'resnames'],
                             trajectory=True)

    @staticmethod
    @pytest.fixture()
    def universe(ref):
        return mda.Universe(ref.topology, ref.trajectory)

    def test_write_different_box(self, ref, universe, tmpdir):
        if ref.changing_dimensions:
            outfile = 'write-dimensions-test' + ref.ext
            with tmpdir.as_cwd():
                with ref.writer(outfile, universe.atoms.n_atoms) as W:
                    for ts in universe.trajectory:
                        universe.dimensions[:3] += 1
                        W.write(universe)

                written = ref.reader(outfile)

                for ts_ref, ts_w in zip(universe.trajectory, written):
                    universe.dimensions[:3] += 1
                    assert_array_almost_equal(universe.dimensions,
                                              ts_w.dimensions,
                                              decimal=ref.prec)

    def test_write_trajectory_atomgroup(self, ref,reader, universe, tmpdir):
        outfile = 'write-atoms-test.' + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(outfile, universe.atoms.n_atoms) as w:
                for ts in universe.trajectory:
                    w.write(universe.atoms)
            self._check_copy(outfile, ref, reader)

    def test_write_trajectory_universe(self, ref, reader, universe, tmpdir):
        outfile = 'write-uni-test.' + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(outfile, universe.atoms.n_atoms) as w:
                for ts in universe.trajectory:
                    w.write(universe)
            self._check_copy(outfile, ref, reader)

    def test_write_selection(self, ref, reader, universe, u_no_resnames,
                             u_no_resids, u_no_names, tmpdir):
        sel_str = 'resid 1'
        sel = universe.select_atoms(sel_str)
        outfile = 'write-selection-test.' + ref.ext


        with tmpdir.as_cwd():
            with ref.writer(outfile, sel.n_atoms) as W:
                for ts in universe.trajectory:
                    W.write(sel.atoms)

            copy = ref.reader(outfile)
            for orig_ts, copy_ts in zip(universe.trajectory, copy):
                assert_array_almost_equal(
                    copy_ts._pos, sel.atoms.positions, ref.prec,
                    err_msg="coordinate mismatch between original and written "
                            "trajectory at frame {} (orig) vs {} (copy)".format(
                        orig_ts.frame, copy_ts.frame))

    def _check_copy(self, fname, ref, reader):
        copy = ref.reader(fname)
        assert_equal(reader.n_frames, copy.n_frames)
        for orig_ts, copy_ts in zip(reader, copy):
            assert_timestep_almost_equal(
                copy_ts, orig_ts, decimal=ref.prec)

    def test_write_none(self, ref, tmpdir):
        outfile = 'write-none.' + ref.ext
        with tmpdir.as_cwd():
            with pytest.raises(TypeError):
                with ref.writer(outfile, 42) as w:
                    w.write(None)

    def test_no_container(self, ref, tmpdir):
        with tmpdir.as_cwd():
            if ref.container_format:
                ref.writer('foo')
            else:
                with pytest.raises(TypeError):
                    ref.writer('foo')

    def test_write_not_changing_ts(self, ref, universe, tmpdir):
        outfile = 'write-not-changing-ts.' + ref.ext

        copy_ts = universe.trajectory.ts.copy()
        with tmpdir.as_cwd():
            with ref.writer(outfile, n_atoms=5) as W:
                W.write(universe)
                assert_timestep_almost_equal(copy_ts, universe.trajectory.ts)


def assert_timestep_equal(A, B, msg=''):
    """ assert that two timesteps are exactly equal and commutative
    """
    assert A == B, msg
    assert B == A, msg


def assert_timestep_almost_equal(A, B, decimal=6, verbose=True):
    if not isinstance(A, Timestep):
        raise AssertionError('A is not of type Timestep')
    if not isinstance(B, Timestep):
        raise AssertionError('B is not of type Timestep')

    if A.frame != B.frame:
        raise AssertionError('A and B refer to different frames: '
                             'A.frame = {}, B.frame={}'.format(
                                 A.frame, B.frame))

    if not np.allclose(A.time, B.time):
        raise AssertionError('A and B refer to different times: '
                             'A.time = {}, B.time={}'.format(
                                 A.time, B.time))

    if A.n_atoms != B.n_atoms:
        raise AssertionError('A and B have a differnent number of atoms: '
                             'A.n_atoms = {}, B.n_atoms = {}'.format(
                                 A.n_atoms, B.n_atoms))

    if A.has_positions != B.has_positions:
        raise AssertionError('Only one Timestep has positions:'
                             'A.has_positions = {}, B.has_positions = {}'.format(
                                 A.has_positions, B.has_positions))

    if A.has_positions:
        assert_array_almost_equal(A.positions, B.positions, decimal=decimal,
                                  err_msg='Timestep positions',
                                  verbose=verbose)

    if A.has_velocities != B.has_velocities:
        raise AssertionError('Only one Timestep has velocities:'
                             'A.has_velocities = {}, B.has_velocities = {}'.format(
                                 A.has_velocities, B.has_velocities))
    if A.has_velocities:
        assert_array_almost_equal(A.velocities, B.velocities, decimal=decimal,
                                  err_msg='Timestep velocities',
                                  verbose=verbose)

    if A.has_forces != B.has_forces:
        raise AssertionError('Only one Timestep has forces:'
                             'A.has_forces = {}, B.has_forces = {}'.format(
                                 A.has_forces, B.has_forces))
    if A.has_forces:
        assert_array_almost_equal(A.forces, B.forces, decimal=decimal,
                                  err_msg='Timestep forces', verbose=verbose)

    # Check we've got auxiliaries before comparing values (auxiliaries aren't written
    # so we won't have aux values to compare when testing writer)
    if len(A.aux) > 0 and len(B.aux) > 0:
        assert_equal(A.aux, B.aux, err_msg='Auxiliary values do not match: '
                                  'A.aux = {}, B.aux = {}'.format(A.aux, B.aux))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import MDAnalysis as mda
from MDAnalysis.analysis.rdf import InterRDF

from MDAnalysisTests.datafiles import two_water_gro


@pytest.fixture()
def u():
    return mda.Universe(two_water_gro, in_memory=True)


@pytest.fixture()
def sels(u):
    # modify the coordinates to produce specific test results
    # (NOTE: requires in-memory coordinates to make them permanent)
    for at, (x, y) in zip(u.atoms, zip([1] * 3 + [2] * 3, [2, 1, 3] * 2)):
        at.position = x, y, 0.0
    s1 = u.select_atoms('name OW')
    s2 = u.select_atoms('name HW1 HW2')
    return s1, s2


def test_nbins(u):
    s1 = u.atoms[:3]
    s2 = u.atoms[3:]
    rdf = InterRDF(s1, s2, nbins=412).run()

    assert len(rdf.results.bins) == 412


def test_range(u):
    s1 = u.atoms[:3]
    s2 = u.atoms[3:]
    rmin, rmax = 1.0, 13.0
    rdf = InterRDF(s1, s2, range=(rmin, rmax)).run()

    assert rdf.results.edges[0] == rmin
    assert rdf.results.edges[-1] == rmax


def test_count_sum(sels):
    # OW vs HW
    # should see 8 comparisons in count
    s1, s2 = sels
    rdf = InterRDF(s1, s2).run()
    assert rdf.results.count.sum() == 8


def test_count(sels):
    # should see two distances with 4 counts each
    s1, s2 = sels
    rdf = InterRDF(s1, s2).run()
    assert len(rdf.results.count[rdf.results.count == 4]) == 2


def test_double_run(sels):
    # running rdf twice should give the same result
    s1, s2 = sels
    rdf = InterRDF(s1, s2).run()
    rdf.run()
    assert len(rdf.results.count[rdf.results.count == 4]) == 2


def test_exclusion(sels):
    # should see two distances with 4 counts each
    s1, s2 = sels
    rdf = InterRDF(s1, s2, exclusion_block=(1, 2)).run()
    assert rdf.results.count.sum() == 4


@pytest.mark.parametrize("attr", ("rdf", "bins", "edges", "count"))
def test_rdf_attr_warning(sels, attr):
    s1, s2 = sels
    rdf = InterRDF(s1, s2).run()
    wmsg = f"The `{attr}` attribute was deprecated in MDAnalysis 2.0.0"
    with pytest.warns(DeprecationWarning, match=wmsg):
        getattr(rdf, attr) is rdf.results[attr]



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from io import StringIO

import pytest
import numpy as np
from numpy.testing import assert_equal
import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    PDB,
    PDB_HOLE,
    PDB_small,
    PDB_conect,
    PDB_conect2TER,
    PDB_singleconect,
    PDB_chainidnewres,
    PDB_sameresid_diffresname,
    PDB_helix,
    PDB_elements,
)
from MDAnalysis.topology.PDBParser import PDBParser
from MDAnalysis import NoDataError

_PDBPARSER = mda.topology.PDBParser.PDBParser

hybrid36 = [
    ("A0000", 100000),
    ("MEGAN", 20929695),
    ("J0NNY", 15247214),
    ("DREW6", 6417862),
    ("ST3V3", 31691119),
    ("ADA8M", 719798),
    ("a0000", 43770016),
    ("megan", 64599711),
    ("j0nny", 58917230),
    ("drew6", 50087878),
    ("st3v3", 75361135),
    ("ada8m", 44389814),
    ("    6", 6),
    ("   24", 24),
    ("  645", 645),
    (" 4951", 4951),
    ("10267", 10267)
]


@pytest.mark.parametrize('hybrid, integer', hybrid36)
def test_hy36decode(hybrid, integer):
    assert mda.topology.PDBParser.hy36decode(5, hybrid) == integer


class PDBBase(ParserBase):
    expected_attrs = ['ids', 'names', 'record_types', 'resids',
                      'resnames', 'altLocs', 'icodes', 'occupancies',
                      'tempfactors', 'chainIDs']
    guessed_attrs = ['types', 'masses']


class TestPDBParser(PDBBase):
    """This one has neither chainids or segids"""
    parser = mda.topology.PDBParser.PDBParser
    ref_filename = PDB
    expected_n_atoms = 47681
    expected_n_residues = 11302
    expected_n_segments = 1


class TestPDBParserSegids(PDBBase):
    """Has segids"""
    parser = mda.topology.PDBParser.PDBParser
    ref_filename = PDB_small
    expected_n_atoms = 3341
    expected_n_residues = 214
    expected_n_segments = 1


class TestPDBConect(object):
    """Testing PDB topology parsing (PDB)"""

    def test_conect_parser(self):
        lines = ("CONECT1233212331",
                 "CONECT123331233112334",
                 "CONECT123341233312335",
                 "CONECT123351233412336",
                 "CONECT12336123271233012335",
                 "CONECT12337 7718 84081234012344",
                 "CONECT1233812339123401234112345")
        results = ((12332, [12331]),
                   (12333, [12331, 12334]),
                   (12334, [12333, 12335]),
                   (12335, [12334, 12336]),
                   (12336, [12327, 12330, 12335]),
                   (12337, [7718, 8408, 12340, 12344]),
                   (12338, [12339, 12340, 12341, 12345]))
        for line, res in zip(lines, results):
            bonds = mda.topology.PDBParser._parse_conect(line)
            assert_equal(bonds[0], res[0])
            for bond, res_bond in zip(bonds[1], res[1]):
                assert_equal(bond, res_bond)

    def test_conect_parser_runtime(self):
        with pytest.raises(RuntimeError):
            mda.topology.PDBParser._parse_conect('CONECT12337 7718 '
                                                 '84081234012344123')

    def test_conect_topo_parser(self):
        """Check that the parser works as intended,
        and that the returned value is a dictionary
        """
        with _PDBPARSER(PDB_conect) as p:
            top = p.parse()
            assert isinstance(top, mda.core.topology.Topology)


def test_conect2ter():
    def parse():
        with PDBParser(PDB_conect2TER) as p:
            struc = p.parse()
        return struc

    with pytest.warns(UserWarning):
        struc = parse()

    assert hasattr(struc, 'bonds')
    assert len(struc.bonds.values) == 4


def test_single_conect():
    def parse():
        with PDBParser(PDB_singleconect) as p:
            struc = p.parse()
        return struc

    with pytest.warns(UserWarning):
        struc = parse()
    assert hasattr(struc, 'bonds')
    assert len(struc.bonds.values) == 2


def test_new_chainid_new_res():
    # parser must start new residue when chainid starts

    u = mda.Universe(PDB_chainidnewres)

    assert len(u.residues) == 4
    assert_equal(u.residues.resids, [1, 2, 3, 3])
    assert len(u.segments) == 4
    assert_equal(u.segments.segids, ['A', 'B', 'C', 'D'])
    assert len(u.segments[0].atoms) == 5
    assert len(u.segments[1].atoms) == 5
    assert len(u.segments[2].atoms) == 5
    assert len(u.segments[3].atoms) == 7


def test_sameresid_diffresname():
    with _PDBPARSER(PDB_sameresid_diffresname) as p:
        top = p.parse()
    resids = [9, 9]
    resnames = ['GLN', 'POPC']
    for i, (resid, resname) in enumerate(zip(resids, resnames)):
        assert top.resids.values[i] == resid
        assert top.resnames.values[i] == resname


def test_PDB_record_types():
    u = mda.Universe(PDB_HOLE)

    assert u.atoms[0].record_type == 'ATOM'
    assert u.atoms[132].record_type == 'HETATM'

    assert_equal(u.atoms[10:20].record_types, 'ATOM')
    assert_equal(u.atoms[271:].record_types, 'HETATM')


PDB_noresid = """\
REMARK For testing reading of CRYST
REMARK This has MODELs then CRYST entries
CRYST1   80.000   80.017   80.017  90.00  90.00  90.00 P 1           1
MODEL        1
ATOM      1  H2  TIP3           10.000  44.891  14.267  1.00  0.00      TIP3
ATOM      2  OH2 TIP3           67.275  48.893  23.568  1.00  0.00      TIP3
ATOM      3  H1  TIP3           66.641  48.181  23.485  1.00  0.00      TIP3
ATOM      4  H2  TIP3           66.986  49.547  22.931  1.00  0.00      TIP3
ENDMDL
"""


def test_PDB_no_resid():
    u = mda.Universe(StringIO(PDB_noresid), format='PDB')

    assert len(u.atoms) == 4
    assert len(u.residues) == 1
    # should have default resid of 1
    assert u.residues[0].resid == 1


PDB_hex = """\
REMARK For testing reading of hex atom numbers
REMARK This has MODELs then hex atom numbers entries
CRYST1   80.000   80.017   80.017  90.00  90.00  90.00 P 1           1
MODEL        1
HETATM    1  H     2 L 400      20.168  00.034  40.428
HETATMA0000  H     2 L 400      40.168  50.034  40.428
HETATMA0001  H     2 L 400      30.453  60.495  50.132
HETATMA0002  H     2 L 400      20.576  40.354  60.483
HETATMA0003  H     2 L 400      10.208  30.067  70.045
END
"""

# this causes an error on Win64/Python 3.8 on Azure when loaded
# in as a file instead
PDB_metals = """\
HETATM    1 CU    CU A   1      00.000  00.000  00.000  1.00 00.00          Cu
HETATM    2 FE    FE A   2      03.000  03.000  03.000  1.00 00.00          Fe
HETATM    3 Ca    Ca A   3      03.000  03.000  03.000  1.00 00.00          Ca
HETATM    3 Mg    Mg A   3      03.000  03.000  03.000  1.00 00.00          Mg
"""


def test_PDB_hex():
    u = mda.Universe(StringIO(PDB_hex), format='PDB')
    assert len(u.atoms) == 5
    assert u.atoms[0].id == 1
    assert u.atoms[1].id == 100000
    assert u.atoms[2].id == 100001
    assert u.atoms[3].id == 100002
    assert u.atoms[4].id == 100003


@pytest.mark.filterwarnings("error:Failed to guess the mass")
def test_PDB_metals():
    from MDAnalysis.topology import tables

    u = mda.Universe(StringIO(PDB_metals), format='PDB')

    assert len(u.atoms) == 4
    assert u.atoms[0].mass == pytest.approx(tables.masses["CU"])
    assert u.atoms[1].mass == pytest.approx(tables.masses["FE"])
    assert u.atoms[2].mass == pytest.approx(tables.masses["CA"])
    assert u.atoms[3].mass == pytest.approx(tables.masses["MG"])


def test_PDB_elements():
    """The test checks whether elements attribute are assigned
    properly given a PDB file with valid elements record.
    """
    u = mda.Universe(PDB_elements, format='PDB')
    element_list = np.array(['N', 'C', 'C', 'O', 'C', 'C', 'O', 'N', 'H',
                             'H', 'H', 'H', 'H', 'H', 'H', 'H', 'Cu', 'Fe',
                             'Mg', 'Ca', 'S', 'O', 'C', 'C', 'S', 'O', 'C',
                             'C'], dtype=object)
    assert_equal(u.atoms.elements, element_list)


def test_missing_elements_noattribute():
    """Check that:

    1) a warning is raised if elements are missing
    2) the elements attribute is not set
    """
    wmsg = ("Element information is missing, elements attribute will not be "
            "populated")
    with pytest.warns(UserWarning, match=wmsg):
        u = mda.Universe(PDB_small)
    with pytest.raises(AttributeError):
        _ = u.atoms.elements


PDB_wrong_ele = """\
REMARK For testing warnings of wrong elements
REMARK This file represent invalid elements in the elements column
ATOM      1  N   ASN A   1      -8.901   4.127  -0.555  1.00  0.00           N
ATOM      2  CA  ASN A   1      -8.608   3.135  -1.618  1.00  0.00
ATOM      3  C   ASN A   1      -7.117   2.964  -1.897  1.00  0.00           C
ATOM      4  O   ASN A   1      -6.634   1.849  -1.758  1.00  0.00           O
ATOM      5  X   ASN A   1      -9.437   3.396  -2.889  1.00  0.00          XX
TER       6
HETATM    7 CU    CU A   2      03.000  00.000  00.000  1.00 00.00          CU
HETATM    8 FE    FE A   3      00.000  03.000  00.000  1.00 00.00          Fe
HETATM    9 Mg    Mg A   4      03.000  03.000  03.000  1.00 00.00          MG
TER       10
"""


def test_wrong_elements_warnings():
    """The test checks whether there are invalid elements in the elements
    column which have been parsed and returns an appropriate warning.
    """
    with pytest.warns(UserWarning, match='Unknown element XX found'):
        u = mda.Universe(StringIO(PDB_wrong_ele), format='PDB')

    expected = np.array(['N', '', 'C', 'O', '', 'Cu', 'Fe', 'Mg'],
                        dtype=object)
    assert_equal(u.atoms.elements, expected)


def test_nobonds_error():
    """Issue #2832: PDB without CONECT record should not have a bonds
    attribute and raises NoDataError on access"""
    u = mda.Universe(PDB_helix)

    errmsg = "This Universe does not contain bonds information"

    with pytest.raises(NoDataError, match=errmsg):
        u.atoms.bonds



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import os
import textwrap
from io import StringIO
import itertools
import numpy as np
from numpy.testing import(
    assert_equal,
)

import MDAnalysis
import MDAnalysis as mda
import MDAnalysis.core.selection
from MDAnalysis.lib.distances import distance_array
from MDAnalysis.core.selection import Parser
from MDAnalysis import SelectionError, SelectionWarning

from MDAnalysis.tests.datafiles import (
    PSF, DCD,
    PRMpbc, TRJpbc_bz2,
    PSF_NAMD, PDB_NAMD,
    GRO, RNA_PSF, NUCLsel, TPR, XTC,
    TRZ_psf, TRZ,
    PDB_icodes,
    PDB_HOLE,
    PDB_helix,
    PDB_elements,
)
from MDAnalysisTests import make_Universe

import pytest


class TestSelectionsCHARMM(object):
    @pytest.fixture(scope='class')
    def universe(self):
        """Set up the standard AdK system in implicit solvent.
        Geometry here is orthogonal
        """
        return MDAnalysis.Universe(PSF, DCD)

    @pytest.fixture()
    def universe_copy(self, universe):
        return MDAnalysis.Universe(PSF, DCD)

    def test_segid(self, universe):
        sel = universe.select_atoms('segid 4AKE')
        assert_equal(sel.n_atoms, 3341, "failed to select segment 4AKE")
        assert_equal(sorted(sel.indices),
                     sorted(universe.select_atoms('segid 4AKE').indices),
                     "selected segment 4AKE is not the same as auto-generated segment s4AKE")

    def test_protein(self, universe):
        sel = universe.select_atoms('protein')
        assert_equal(sel.n_atoms, 3341, "failed to select protein")
        assert_equal(sorted(sel.indices),
                     sorted(universe.select_atoms('segid 4AKE').indices),
                     "selected protein is not the same as auto-generated protein segment s4AKE")

    @pytest.mark.parametrize('resname', sorted(MDAnalysis.core.selection.ProteinSelection.prot_res))
    def test_protein_resnames(self, resname):
        u = make_Universe(('resnames',))
        # set half the residues' names to the resname we're testing
        myprot = u.residues[::2]
        # Windows note: the parametrized test input string objects
        # are actually of type 'numpy.str_' and coercion to str
        # proper is needed for unit test on Windows
        myprot.resnames = str(resname)
        # select protein
        sel = u.select_atoms('protein')
        # check that contents (atom indices) are identical afterwards
        assert_equal(myprot.atoms.ix, sel.ix)

    def test_backbone(self, universe):
        sel = universe.select_atoms('backbone')
        assert_equal(sel.n_atoms, 855)

    def test_resid_single(self, universe):
        sel = universe.select_atoms('resid 100')
        assert_equal(sel.n_atoms, 7)
        assert_equal(sel.residues.resnames, ['GLY'])

    def test_resid_range(self, universe):
        sel = universe.select_atoms('resid 100:105')
        assert_equal(sel.n_atoms, 89)
        assert_equal(sel.residues.resnames,
                     ['GLY', 'ILE', 'ASN', 'VAL', 'ASP', 'TYR'])

    def test_selgroup(self, universe):
        sel = universe.select_atoms('not resid 100')
        sel2 = universe.select_atoms('not group notr100', notr100=sel)
        assert_equal(sel2.n_atoms, 7)
        assert_equal(sel2.residues.resnames, ['GLY'])

    def test_fullselgroup(self, universe):
        sel1 = universe.select_atoms('resid 101')
        sel2 = universe.select_atoms('resid 100')
        sel3 = sel1.select_atoms('global group r100', r100=sel2)
        assert_equal(sel2.n_atoms, 7)
        assert_equal(sel2.residues.resnames, ['GLY'])

    # resnum selections are boring here because we haven't really a mechanism yet
    # to assign the canonical PDB resnums
    def test_resnum_single(self, universe):
        sel = universe.select_atoms('resnum 100')
        assert_equal(sel.n_atoms, 7)
        assert_equal(sel.residues.resids, [100])
        assert_equal(sel.residues.resnames, ['GLY'])

    def test_resnum_range(self, universe):
        sel = universe.select_atoms('resnum 100:105')
        assert_equal(sel.n_atoms, 89)
        assert_equal(sel.residues.resids, range(100, 106))
        assert_equal(sel.residues.resnames,
                     ['GLY', 'ILE', 'ASN', 'VAL', 'ASP', 'TYR'])

    def test_resname(self, universe):
        sel = universe.select_atoms('resname LEU')
        assert_equal(sel.n_atoms, 304,
                     "Failed to find all 'resname LEU' atoms.")
        assert_equal(sel.n_residues, 16,
                     "Failed to find all 'resname LEU' residues.")
        assert_equal(sorted(sel.indices),
                     sorted(universe.select_atoms('segid 4AKE and resname LEU').indices),
                     "selected 'resname LEU' atoms are not the same as auto-generated s4AKE.LEU")

    def test_name(self, universe):
        sel = universe.select_atoms('name CA')
        assert_equal(sel.n_atoms, 214)

    def test_atom(self, universe):
        sel = universe.select_atoms('atom 4AKE 100 CA')
        assert_equal(len(sel), 1)
        assert_equal(sel.resnames, ['GLY'])
        assert_equal(
            sel.positions,
            np.array([[20.38685226, -3.44224262, -5.92158318]],
                     dtype=np.float32))

    def test_atom_empty(self, universe):
        sel = universe.select_atoms('atom 4AKE 100 XX')  # Does not exist
        assert_equal(len(sel), 0)

    def test_type(self, universe):
        sel = universe.select_atoms("type 1")
        assert_equal(len(sel), 253)

    def test_and(self, universe):
        sel = universe.select_atoms('resname GLY and resid 100')
        assert_equal(len(sel), 7)

    def test_or(self, universe):
        sel = universe.select_atoms('resname LYS or resname ARG')
        assert_equal(sel.n_residues, 31)

    def test_not(self, universe):
        sel = universe.select_atoms('not backbone')
        assert_equal(len(sel), 2486)

    @pytest.mark.parametrize('selstr', [
        'around 4.0 bynum 1943', 
        'around 4.0 index 1942'
    ])
    def test_around(self, universe, selstr):
        sel = universe.select_atoms(selstr)
        assert_equal(len(sel), 32)

    @pytest.mark.parametrize('selstr', [
        'sphlayer 4.0 6.0 bynum 1281',
        'sphlayer 4.0 6.0 index 1280'
    ])
    def test_sphlayer(self, universe, selstr):
        sel = universe.select_atoms(selstr)
        assert_equal(len(sel), 66)

    @pytest.mark.parametrize('selstr', [
        'sphzone 6.0 bynum 1281',
        'sphzone 6.0 index 1280'
    ])
    def test_sphzone(self, universe, selstr):
        sel = universe.select_atoms(selstr)
        assert_equal(len(sel), 86)

    @pytest.mark.parametrize('selstr', [
        'cylayer 4.0 6.0 10 -10 bynum 1281',
        'cylayer 4.0 6.0 10 -10 index 1280'
    ])
    def test_cylayer(self, universe, selstr):
        sel = universe.select_atoms(selstr)
        assert_equal(len(sel), 88)

    def test_empty_cylayer(self, universe):
        empty = universe.select_atoms('cylayer 4.0 6.0 10 -10 name NOT_A_NAME')
        assert_equal(len(empty), 0)

    @pytest.mark.parametrize('selstr', [
        'cyzone 6.0 10 -10 bynum 1281',
        'cyzone 6.0 10 -10 index 1280'
    ])
    def test_cyzone(self, universe, selstr):
        sel = universe.select_atoms(selstr)
        assert_equal(len(sel), 166)

    def test_empty_cyzone(self, universe):
        empty = universe.select_atoms('cyzone 6.0 10 -10 name NOT_A_NAME')
        assert_equal(len(empty), 0)

    def test_point(self, universe):
        ag = universe.select_atoms('point 5.0 5.0 5.0 3.5')

        d = distance_array(np.array([[5.0, 5.0, 5.0]], dtype=np.float32),
                           universe.atoms.positions,
                           box=universe.dimensions)

        idx = np.where(d < 3.5)[1]

        assert_equal(set(ag.indices), set(idx))

    def test_prop(self, universe):
        sel = universe.select_atoms('prop y <= 16')
        sel2 = universe.select_atoms('prop abs z < 8')
        assert_equal(len(sel), 3194)
        assert_equal(len(sel2), 2001)

    def test_bynum(self, universe):
        "Tests the bynum selection, also from AtomGroup instances (Issue 275)"
        sel = universe.select_atoms('bynum 5')
        assert_equal(sel[0].index, 4)
        sel = universe.select_atoms('bynum 1:10')
        assert_equal(len(sel), 10)
        assert_equal(sel[0].index, 0)
        assert_equal(sel[-1].index, 9)
        subsel = sel.select_atoms('bynum 5')
        assert_equal(subsel[0].index, 4)
        subsel = sel.select_atoms('bynum 2:5')
        assert_equal(len(subsel), 4)
        assert_equal(subsel[0].index, 1)
        assert_equal(subsel[-1].index, 4)

    def test_index(self, universe):
        "Tests the index selection, also from AtomGroup instances (Issue 275)"
        sel = universe.select_atoms('index 4')
        assert_equal(sel[0].index, 4)
        sel = universe.select_atoms('index 0:9')
        assert_equal(len(sel), 10)
        assert_equal(sel[0].index, 0)
        assert_equal(sel[-1].index, 9)
        subsel = sel.select_atoms('index 4')
        assert_equal(subsel[0].index, 4)
        subsel = sel.select_atoms('index 1:4')
        assert_equal(len(subsel), 4)
        assert_equal(subsel[0].index, 1)
        assert_equal(subsel[-1].index, 4)

    @pytest.mark.parametrize('selstr,expected', (
        ['byres bynum 0:5', 19],
        ['byres index 0:19', 43]
    ))
    def test_byres(self, universe, selstr, expected):
        sel = universe.select_atoms(selstr)
        assert_equal(len(sel), expected)

    def test_same_resname(self, universe):
        """Test the 'same ... as' construct (Issue 217)"""
        sel = universe.select_atoms("same resname as resid 10 or resid 11")
        assert_equal(len(sel), 331,
                     ("Found a wrong number of atoms with same resname as "
                      "resids 10 or 11"))
        target_resids = np.array([7, 8, 10, 11, 12, 14, 17, 25, 32, 37, 38,
                                  42, 46, 49, 55, 56, 66, 73, 80, 85, 93, 95,
                                  99, 100, 122, 127, 130, 144, 150, 176, 180,
                                  186, 188, 189, 194, 198, 203, 207, 214])
        assert_equal(sel.residues.resids, target_resids,
                     ("Found wrong residues with same resname as "
                      "resids 10 or 11"))

    def test_same_segment(self, universe_copy):
        """Test the 'same ... as' construct (Issue 217)"""

        SNew_A = universe_copy.add_Segment(segid='A')
        SNew_B = universe_copy.add_Segment(segid='B')
        SNew_C = universe_copy.add_Segment(segid='C')

        universe_copy.residues[:100].segments = SNew_A
        universe_copy.residues[100:150].segments = SNew_B
        universe_copy.residues[150:].segments = SNew_C

        target_resids = np.arange(100) + 1
        sel = universe_copy.select_atoms("same segment as resid 10")
        assert_equal(len(sel), 1520,
                     "Found a wrong number of atoms in the same segment of resid 10")
        assert_equal(sel.residues.resids,
                     target_resids,
                     "Found wrong residues in the same segment of resid 10")

        target_resids = np.arange(100, 150) + 1
        sel = universe_copy.select_atoms("same segment as resid 110")
        assert_equal(len(sel), 797,
                     "Found a wrong number of atoms in the same segment of resid 110")
        assert_equal(sel.residues.resids, target_resids,
                     "Found wrong residues in the same segment of resid 110")

        target_resids = np.arange(150, universe_copy.atoms.n_residues) + 1
        sel = universe_copy.select_atoms("same segment as resid 160")
        assert_equal(len(sel), 1024,
                     "Found a wrong number of atoms in the same segment of resid 160")
        assert_equal(sel.residues.resids, target_resids,
                     "Found wrong residues in the same segment of resid 160")

    def test_empty_same(self, universe):
        ag = universe.select_atoms('resname MET')

        # No GLY, so 'as resname GLY' is empty
        ag2 = ag.select_atoms('same mass as resname GLY')

        assert len(ag2) == 0

    def test_empty_selection(self, universe):
        """Test that empty selection can be processed (see Issue 12)"""
        # no Trp in AdK
        assert_equal(len(universe.select_atoms('resname TRP')), 0)

    def test_parenthesized_expression(self, universe):
        sel = universe.select_atoms(
            '( name CA or name CB ) and resname LEU')
        assert_equal(len(sel), 32)

    def test_no_space_around_parentheses(self, universe):
        """Test that no space is needed around parentheses (Issue 43)."""
        # note: will currently be ERROR because it throws a ParseError
        sel = universe.select_atoms('(name CA or name CB) and resname LEU')
        assert_equal(len(sel), 32)

    # TODO:
    # test for checking ordering and multiple comma-separated selections
    def test_concatenated_selection(self, universe):
        E151 = universe.select_atoms('segid 4AKE').select_atoms('resid 151')
        # note that this is not quite phi... HN should be C of prec. residue
        phi151 = E151.atoms.select_atoms('name HN', 'name N', 'name CA',
                                         'name CB')
        assert_equal(len(phi151), 4)
        assert_equal(phi151[0].name, 'HN',
                     "wrong ordering in selection, should be HN-N-CA-CB")

    def test_global(self, universe):
        """Test the `global` modifier keyword (Issue 268)"""
        ag = universe.select_atoms("resname LYS and name NZ")
        # Lys amines within 4 angstrom of the backbone.
        ag1 = universe.select_atoms(
            "resname LYS and name NZ and around 4 backbone")
        ag2 = ag.select_atoms("around 4 global backbone")
        assert_equal(ag2.indices, ag1.indices)

    @pytest.mark.parametrize('selstring, wildstring', [
        ('resname TYR THR', 'resname T*R'),
        ('resname ASN GLN', 'resname *N'),
        ('resname ASN ASP', 'resname AS*'),
        ('resname TYR THR', 'resname T?R'),
        ('resname ASN ASP HSD', 'resname *S?'),
        ('resname LEU LYS', 'resname L**'),
        ('resname MET', 'resname *M*'),
        ('resname GLN GLY', 'resname GL[NY]'),
        ('resname GLU', 'resname GL[!NY]'),
    ])
    def test_wildcard_selection(self, universe, selstring, wildstring):
        ag = universe.select_atoms(selstring)
        ag_wild = universe.select_atoms(wildstring)
        assert ag == ag_wild

class TestSelectionsAMBER(object):
    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(PRMpbc, TRJpbc_bz2)

    def test_protein(self, universe):
        sel = universe.select_atoms('protein')
        assert_equal(sel.n_atoms, 22, "failed to select protein")

    def test_backbone(self, universe):
        sel = universe.select_atoms('backbone')
        assert_equal(sel.n_atoms, 7)

    def test_type(self, universe):
        sel = universe.select_atoms('type HC')
        assert_equal(len(sel), 6)
        assert_equal(sel.names, ['HH31', 'HH32', 'HH33', 'HB1', 'HB2', 'HB3'])


class TestSelectionsNAMD(object):
    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(PSF_NAMD, PDB_NAMD)

    def test_protein(self, universe):
        # must include non-standard residues
        sel = universe.select_atoms(
            'protein or resname HAO or resname ORT')
        assert_equal(sel.n_atoms, universe.atoms.n_atoms,
                     "failed to select peptide")
        assert_equal(sel.n_residues, 6,
                     "failed to select all peptide residues")

    def test_resid_single(self, universe):
        sel = universe.select_atoms('resid 12')
        assert_equal(sel.n_atoms, 26)
        assert_equal(sel.residues.resnames, ['HAO'])

    def test_type(self, universe):
        sel = universe.select_atoms('type H')
        assert_equal(len(sel), 5)
        # note 4th HH
        assert_equal(sel.names, ['HN', 'HN', 'HN', 'HH', 'HN'])


class TestSelectionsGRO(object):
    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(GRO)

    def test_protein(self, universe):
        sel = universe.select_atoms('protein')
        assert_equal(sel.n_atoms, 3341, "failed to select protein")

    def test_backbone(self, universe):
        sel = universe.select_atoms('backbone')
        assert_equal(sel.n_atoms, 855)

    def test_resid_single(self, universe):
        sel = universe.select_atoms('resid 100')
        assert_equal(sel.n_atoms, 7)
        assert_equal(sel.residues.resnames, ['GLY'])

    @pytest.mark.parametrize('selstr', [
        'same x as bynum 1 or bynum 10',
        'same x as bynum 1 10',
        'same x as index 0 or index 9',
        'same x as index 0 9'
    ])
    def test_same_coordinate(self, universe, selstr):
        """Test the 'same ... as' construct (Issue 217)"""
        sel = universe.select_atoms(selstr)
        errmsg = ("Found a wrong number of atoms with same "
                  "x as ids 1 or 10")
        assert_equal(len(sel), 12, errmsg) 
        target_ids = np.array([0, 8, 9, 224, 643, 3515, 11210, 14121,
                               18430, 25418, 35811, 43618])
        assert_equal(sel.indices, target_ids,
                     "Found wrong atoms with same x as ids 1 or 10")

    @pytest.mark.parametrize('selstr', [
        'cylayer 10 20 20 -20 bynum 3554',
        'cylayer 10 20 20 -20 index 3553'
    ])
    def test_cylayer(self, universe, selstr):
        """Cylinder layer selections with tricilinic periodicity (Issue 274)"""
        atgp = universe.select_atoms('name OW')
        sel = atgp.select_atoms(selstr)
        assert_equal(len(sel), 1155)

    @pytest.mark.parametrize('selstr', [
        'cyzone 20 20 -20 bynum 3554',
        'cyzone 20 20 -20 index 3553'
    ])
    def test_cyzone(self, universe, selstr):
        """Cylinder zone selections with tricilinic periodicity (Issue 274)"""
        atgp = universe.select_atoms('name OW')
        sel = atgp.select_atoms(selstr)
        assert_equal(len(sel), 1556)


class TestSelectionsTPR(object):
    @staticmethod
    @pytest.fixture(scope='class')
    def universe():
        return MDAnalysis.Universe(TPR, XTC, tpr_resid_from_one=False)

    @pytest.mark.parametrize('selstr', [
        'same fragment as bynum 1',
        'same fragment as index 0'
    ])
    def test_same_fragment(self, universe, selstr):
        """Test the 'same ... as' construct (Issue 217)"""
        # This test comes here because it's a system with solvent,
        # and thus multiple fragments.
        sel = universe.select_atoms(selstr)
        errmsg = ("Found a wrong number of atoms "
                  "on the same fragment as id 1")
        assert_equal(len(sel), 3341, errmsg)
        errmsg = ("Found a differ set of atoms when using the 'same "
                  "fragment as' construct vs. the .fragment property")
        assert_equal(sel.indices, universe.atoms[0].fragment.indices, errmsg)

    def test_moltype(self, universe):
        sel = universe.select_atoms("moltype NA+")
        ref = np.array([47677, 47678, 47679, 47680], dtype=np.int32)
        assert_equal(sel.ids, ref)

    @pytest.mark.parametrize(
        'selection_string,reference',
        (('molnum 1', [3341, 3342, 3343, 3344]),
         ('molnum 2:4', [3345, 3346, 3347, 3348, 3349, 3350,
                         3351, 3352, 3353, 3354, 3355, 3356]),
         )
    )
    def test_molnum(self, universe, selection_string, reference):
        sel = universe.select_atoms(selection_string)
        assert_equal(sel.ids, np.array(reference, dtype=np.int32))


class TestSelectionRDKit(object):
    def setup_class(self):
        pytest.importorskip("rdkit.Chem")

    @pytest.fixture
    def u(self):
        smi = "Cc1cNcc1"
        u = MDAnalysis.Universe.from_smiles(smi, addHs=False,
                                            generate_coordinates=False)
        return u

    @pytest.fixture
    def u2(self):
        u = MDAnalysis.Universe.from_smiles("Nc1cc(C[C@H]([O-])C=O)c[nH]1")
        return u

    @pytest.mark.parametrize("sel_str, n_atoms", [
        ("aromatic", 5),
        ("not aromatic", 1),
        ("type N and aromatic", 1),
        ("type C and aromatic", 4),
    ])
    def test_aromatic_selection(self, u, sel_str, n_atoms):
        sel = u.select_atoms(sel_str)
        assert sel.n_atoms == n_atoms

    @pytest.mark.parametrize("sel_str, indices", [
        ("smarts n", [10]),
        ("smarts [#7]", [0, 10]),
        ("smarts a", [1, 2, 3, 9, 10]),
        ("smarts c", [1, 2, 3, 9]),
        ("smarts [*-]", [6]),
        ("smarts [$([!#1]);$([!R][R])]", [0, 4]),
        ("smarts [$([C@H](-[CH2])(-[O-])-C=O)]", [5]),
        ("smarts [$([C@@H](-[CH2])(-[O-])-C=O)]", []),
        ("smarts a and type C", [1, 2, 3, 9]),
        ("(smarts a) and (type C)", [1, 2, 3, 9]),
        ("smarts a and type N", [10]),
    ])
    def test_smarts_selection(self, u2, sel_str, indices):
        sel = u2.select_atoms(sel_str)
        assert_equal(sel.indices, indices)

    def test_invalid_smarts_sel_raises_error(self, u2):
        with pytest.raises(ValueError, match="not a valid SMARTS"):
            u2.select_atoms("smarts foo")

    def test_passing_rdkit_kwargs_to_converter(self):
        u = mda.Universe.from_smiles("O=C=O")
        sel = u.select_atoms("smarts [$(O=C)]", rdkit_kwargs=dict(force=True))
        assert sel.n_atoms == 2

    def test_passing_max_matches_to_converter(self, u2):
        with pytest.warns(UserWarning, match="Your smarts-based") as wsmg:
            sel = u2.select_atoms("smarts C", smarts_kwargs=dict(maxMatches=2))
            sel2 = u2.select_atoms(
                    "smarts C", smarts_kwargs=dict(maxMatches=1000))
            assert sel.n_atoms == 2
            assert sel2.n_atoms == 3

        sel3 = u2.select_atoms("smarts c")
        assert sel3.n_atoms == 4

    def test_passing_use_chirality_to_converter(self):
        u = mda.Universe.from_smiles("CC[C@H](C)O")
        sel3 = u.select_atoms("byres smarts CC[C@@H](C)O")
        assert sel3.n_atoms == 0
        sel4 = u.select_atoms("byres smarts CC[C@@H](C)O", smarts_kwargs={"useChirality": False})
        assert sel4.n_atoms == 15


class TestSelectionsNucleicAcids(object):
    @pytest.fixture(scope='class')
    def universe(self):
        return MDAnalysis.Universe(RNA_PSF)

    @pytest.fixture(scope='class')
    def universe2(self):
        return MDAnalysis.Universe(NUCLsel)


    def test_nucleic(self, universe):
        rna = universe.select_atoms("nucleic")
        assert_equal(rna.n_atoms, 739)
        assert_equal(rna.n_residues, 23)

    def test_nucleic_all(self, universe2):
        sel = universe2.select_atoms('nucleic')
        assert len(sel) == 34

    def test_nucleicbackbone(self, universe):
        rna = universe.select_atoms("nucleicbackbone")
        assert_equal(rna.n_residues, 23)
        assert_equal(rna.n_atoms, rna.n_residues * 5 - 1)
        # -1 because this is a single strand of RNA and on P is missing at the 5' end

    # todo: need checks for other selection resnames such as DT DA DG DC DU

    def test_nucleicbase(self, universe):
        rna = universe.select_atoms("nucleicbase")
        assert_equal(rna.n_residues, 23)
        assert_equal(rna.n_atoms, 214)

    def test_nucleicsugar(self, universe):
        rna = universe.select_atoms("nucleicsugar")
        assert_equal(rna.n_residues, 23)
        assert_equal(rna.n_atoms, rna.n_residues * 5)


class BaseDistanceSelection(object):
    """Both KDTree and distmat selections on orthogonal system

    Selections to check:
     - Around
     - SphericalLayer
     - SphericalZone
     - Point

    Cylindrical methods don't use KDTree
    """
    @pytest.mark.parametrize('periodic', (True, False))
    def test_around(self, u, periodic):
        sel = Parser.parse('around 5.0 resid 1', u.atoms)
        if periodic:
            sel.periodic = True
        else:
            sel.periodic = False

        result = sel.apply(u.atoms)

        r1 = u.select_atoms('resid 1')
        cog = r1.center_of_geometry().reshape(1, 3)

        box = u.dimensions if periodic else None
        d = distance_array(u.atoms.positions, r1.positions,
                           box=box)
        ref = set(np.where(d < 5.0)[0])

        # Around doesn't include atoms from the reference group
        ref.difference_update(set(r1.indices))
        assert ref == set(result.indices)

    @pytest.mark.parametrize('periodic', (True, False))
    def test_spherical_layer(self, u, periodic):
        sel = Parser.parse('sphlayer 2.4 6.0 resid 1', u.atoms)
        if periodic:
            sel.periodic = True
        else:
            sel.periodic = False

        result = sel.apply(u.atoms)

        r1 = u.select_atoms('resid 1')
        box = u.dimensions if periodic else None
        cog = r1.center_of_geometry().reshape(1, 3)
        d = distance_array(u.atoms.positions, cog, box=box)
        ref = set(np.where((d > 2.4) & (d < 6.0))[0])

        assert ref == set(result.indices)

    @pytest.mark.parametrize('periodic', (True, False))
    def test_spherical_zone(self, u, periodic):
        sel = Parser.parse('sphzone 5.0 resid 1', u.atoms)
        if periodic:
            sel.periodic = True
        else:
            sel.periodic = False

        result = sel.apply(u.atoms)

        r1 = u.select_atoms('resid 1')
        box = u.dimensions if periodic else None
        cog = r1.center_of_geometry().reshape(1, 3)
        d = distance_array(u.atoms.positions, cog, box=box)
        ref = set(np.where(d < 5.0)[0])

        assert ref == set(result.indices)

    @pytest.mark.parametrize('periodic', (True, False))
    def test_point(self, u, periodic):
        sel = Parser.parse('point 5.0 5.0 5.0  3.0', u.atoms)
        if periodic:
            sel.periodic = True
        else:
            sel.periodic = False
        result = sel.apply(u.atoms)

        box = u.dimensions if periodic else None
        d = distance_array(np.array([[5.0, 5.0, 5.0]], dtype=np.float32),
                           u.atoms.positions,
                           box=box)
        ref = set(np.where(d < 3.0)[1])

        assert ref == set(result.indices)


class TestOrthogonalDistanceSelections(BaseDistanceSelection):
    @pytest.fixture()
    def u(self):
        return mda.Universe(TRZ_psf, TRZ)

    @pytest.mark.parametrize('meth, periodic', [
        ('distmat', True),
        ('distmat', False)
    ])
    def test_cyzone(self, u, meth, periodic):
        sel = Parser.parse('cyzone 5 4 -4 resid 2', u.atoms)
        sel.periodic = periodic
        result = sel.apply(u.atoms)

        other = u.select_atoms('resid 2')
        pos = other.center_of_geometry()

        vecs = u.atoms.positions - pos
        if periodic:
            box = u.dimensions[:3]
            vecs -= box * np.rint(vecs / box)

        mask = (vecs[:, 2] > -4) & (vecs[:, 2] < 4)

        radii = vecs[:, 0] ** 2 + vecs[:, 1] ** 2
        mask &= radii < 5 ** 2

        ref = set(u.atoms[mask].indices)

        assert ref == set(result.indices)

    @pytest.mark.parametrize('periodic,expected', ([True, 33], [False, 25]))
    def test_sphzone(self, u, periodic, expected):
        sel = u.select_atoms('sphzone 5.0 resid 1', periodic=periodic)

        assert len(sel) == expected


class TestTriclinicDistanceSelections(BaseDistanceSelection):
    @pytest.fixture()
    def u(self):
        return mda.Universe(GRO)


class TestTriclinicSelections(object):
    """Non-KDTree based selections

    This system has triclinic geometry so won't use KDTree based selections
    """

    @pytest.fixture()
    def u(self):
        return mda.Universe(GRO)

    def test_around(self, u):
        r1 = u.select_atoms('resid 1')

        ag = u.select_atoms('around 5.0 resid 1')

        d = distance_array(u.atoms.positions, r1.positions, box=u.dimensions)
        idx = set(np.where(d < 5.0)[0])

        # Around doesn't include atoms from the reference group
        idx.difference_update(set(r1.indices))
        assert idx == set(ag.indices)

    def test_sphlayer(self, u):
        r1 = u.select_atoms('resid 1')
        cog = r1.center_of_geometry().reshape(1, 3)

        ag = u.select_atoms('sphlayer 2.4 6.0 resid 1')

        d = distance_array(u.atoms.positions, cog, box=u.dimensions)
        idx = set(np.where((d > 2.4) & (d < 6.0))[0])

        assert idx == set(ag.indices)

    def test_empty_sphlayer(self, u):
        empty = u.select_atoms('sphlayer 2.4 6.0 name NOT_A_NAME')
        assert len(empty) == 0

    def test_sphzone(self, u):
        r1 = u.select_atoms('resid 1')
        cog = r1.center_of_geometry().reshape(1, 3)

        ag = u.select_atoms('sphzone 5.0 resid 1')

        d = distance_array(u.atoms.positions, cog, box=u.dimensions)
        idx = set(np.where(d < 5.0)[0])

        assert idx == set(ag.indices)

    def test_empty_sphzone(self, u):
        empty = u.select_atoms('sphzone 5.0 name NOT_A_NAME')
        assert len(empty) == 0

    def test_point_1(self, u):
        # The example selection
        ag = u.select_atoms('point 5.0 5.0 5.0 3.5')

        d = distance_array(np.array([[5.0, 5.0, 5.0]], dtype=np.float32),
                           u.atoms.positions,
                           box=u.dimensions)

        idx = np.where(d < 3.5)[1]

        assert_equal(set(ag.indices), set(idx))

    def test_point_2(self, u):
        ag1 = u.atoms[:10000]

        ag2 = ag1.select_atoms('point 5.0 5.0 5.0 3.5')

        d = distance_array(np.array([[5.0, 5.0, 5.0]], dtype=np.float32),
                           ag1.positions,
                           box=u.dimensions)

        idx = np.where(d < 3.5)[1]

        assert_equal(set(ag2.indices), set(idx))


def gen_sel_strings(prop, oper):
    """Generate all possible combinations of spaces in selection strings

    ie:
      'prop x < 1.5'
      'prop x< 1.5'
      'prop x <1.5'
      'prop x<1.5'

    """
    for x, y in itertools.product([' ', ''], [' ', '']):
        yield 'prop {prop}{spc1}{oper}{spc2}1.5'.format(
            prop=prop, spc1=x, oper=oper, spc2=y)


class TestPropSelection(object):
    plurals = {'mass': 'masses',
               'charge': 'charges'}
    op_funcs = {
        '<': np.less,
        '<=': np.less_equal,
        '>': np.greater,
        '>=': np.greater_equal,
        '==': np.equal,
        '!=': np.not_equal
    }
    opposites = {
        '==': '==', '!=': '!=',
        '>': '<=', '<=': '>',
        '<': '>=', '>=': '<',
    }

    @pytest.fixture(params=[slice(None, None), slice(None, 100)])
    def ag(self, request):
        u = make_Universe(('masses', 'charges'))
        u.atoms[::2].masses = 1.5
        u.atoms[::2].charges = 1.5
        return u.atoms[request.param]

    @pytest.mark.parametrize('prop, selstr', [
        (prop, sel)
        for prop in ['mass', 'charge']
        for sel in gen_sel_strings(prop, '<')
    ])
    def test_lt(self, prop, selstr, ag):
        sel = ag.select_atoms(selstr)
        assert_equal(set(sel.indices),
                     set(ag[getattr(ag, self.plurals[prop]) < 1.5].indices))

    @pytest.mark.parametrize('prop, selstr', [
        (prop, sel)
        for prop in ['mass', 'charge']
        for sel in gen_sel_strings(prop, '<=')
    ])
    def test_le(self, prop, selstr, ag):
        sel = ag.select_atoms(selstr)
        assert_equal(set(sel.indices),
                     set(ag[getattr(ag,
                                    self.plurals[prop]) <= 1.5].indices))

    @pytest.mark.parametrize('prop, selstr', [
        (prop, sel)
        for prop in ['mass', 'charge']
        for sel in gen_sel_strings(prop, '>')
    ])
    def test_gt(self, prop, selstr, ag):
        sel = ag.select_atoms(selstr)
        assert_equal(set(sel.indices),
                     set(ag[getattr(ag, self.plurals[prop]) > 1.5].indices))

    @pytest.mark.parametrize('prop, selstr', [
        (prop, sel)
        for prop in ['mass', 'charge']
        for sel in gen_sel_strings(prop, '>=')
    ])
    def test_ge(self, prop, selstr, ag):
        sel = ag.select_atoms(selstr)
        assert_equal(set(sel.indices),
                     set(ag[getattr(ag,
                                    self.plurals[prop]) >= 1.5].indices))

    @pytest.mark.parametrize('prop, selstr', [
        (prop, sel)
        for prop in ['mass', 'charge']
        for sel in gen_sel_strings(prop, '==')
    ])
    def test_eq(self, prop, selstr, ag):
        sel = ag.select_atoms(selstr)
        assert_equal(set(sel.indices),
                     set(ag[getattr(ag,
                                    self.plurals[prop]) == 1.5].indices))

    @pytest.mark.parametrize('prop, selstr', [
        (prop, sel)
        for prop in ['mass', 'charge']
        for sel in gen_sel_strings(prop, '!=')
    ])
    def test_ne(self, prop, selstr, ag):
        sel = ag.select_atoms(selstr)
        assert_equal(set(sel.indices),
                     set(ag[getattr(ag,
                                    self.plurals[prop]) != 1.5].indices))

    @pytest.mark.parametrize('prop, op', [
        (prop, op)
        for prop in ['mass', 'charge']
        for op in ('<', '>', '<=', '>=', '==', '!=')
    ])
    def test_flip(self, prop, ag, op):
        func = self.op_funcs[op]

        # reference group, doing things forwards
        ref = ag[func(getattr(ag, self.plurals[prop]), 1.5)]

        selstr = 'prop 1.5 {op} {prop}'.format(
            op=self.opposites[op], prop=prop)
        sel = ag.select_atoms(selstr)

        assert_equal(set(ref.indices), set(sel.indices))


class TestBondedSelection(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(PSF, DCD)

    def test_bonded_1(self, u):
        ag = u.select_atoms('type 2 and bonded name N')

        assert len(ag) == 3

    def test_nobonds_warns(self, u):
        u = make_Universe(('names',))

        # empty bond topology attr
        batt = mda.core.topologyattrs.Bonds([])
        u.add_TopologyAttr(batt)

        with pytest.warns(UserWarning):
            u.select_atoms('bonded name AAA')


class TestSelectionErrors(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return make_Universe(
            ('names', 'masses', 'resids', 'resnames', 'resnums'))

    @pytest.mark.parametrize('selstr', [
        'name and H',  # string selection
        'name )',
        'resid abcd',  # resid arg parsing selection
        'resnum 7a7',  # rangeselection arg parsing
        'resid 1-',
        'prop chicken == tasty',
        'prop chicken <= 7.4',
        'prop mass ^^ 12.0',
        'same this as resid 1',  # same selection
        'same resid resname mass 5.0',  # same / expect
        'name H and',  # check all tokens used
        'naem H',  # unkonwn (misplet) opertaor
        'resid and name C',  # rangesel not finding vals
        'resnum ',
        'bynum or protein',
        'index or protein',
        'prop mass < 4.0 hello',  # unused token
        'prop mass > 10. and group this',  # missing group
        # bad ranges
        'mass 1.0 to',
        'mass to 3.0',
        'mass 1.0:',
        'mass :3.0',
        'mass 1-',
        'chirality ',
    ])
    def test_selection_fail(self, selstr, universe):
        with pytest.raises(SelectionError):
            universe.select_atoms(selstr)

    def test_invalid_prop_selection(self, universe):
        with pytest.raises(SelectionError, match="Expected one of"):
            universe.select_atoms("prop parsnip < 2")


def test_segid_and_resid():
    u = make_Universe(('segids', 'resids'))

    ag = u.select_atoms('segid SegB and resid 1-100')

    ref = ag.select_atoms('segid SegB').select_atoms('resid 1-100')

    assert_equal(ag.indices, ref.indices)


class TestImplicitOr(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return make_Universe(
            ('names', 'types', 'resids', 'resnums', 'resnames', 'segids'))

    def _check_sels(self, ref, sel, universe):
        ref = universe.select_atoms(ref)
        sel = universe.select_atoms(sel)

        assert_equal(ref.indices, sel.indices)

    @pytest.mark.parametrize('ref, sel', [
        ('name NameABA or name NameACA or name NameADA',
         'name NameABA NameACA NameADA'),
        ('type TypeE or type TypeD or type TypeB', 'type TypeE TypeD TypeB'),
        ('resname RsC or resname RsY', 'resname RsC RsY'),
        ('name NameAB* or name NameACC', 'name NameAB* NameACC'),
        ('segid SegA or segid SegC', 'segid SegA SegC'),
        ('(name NameABC or name NameABB) and (resname RsD or resname RsF)',
         'name NameABC NameABB and resname RsD RsF'),
    ])
    def test_string_selections(self, ref, sel, universe):
        self._check_sels(ref, sel, universe)

    @pytest.mark.parametrize("seltype", ['resid', 'resnum', 'bynum', 'index'])
    @pytest.mark.parametrize('ref, sel', [
        ('{typ} 1 or {typ} 2', '{typ} 1 2'),
        ('{typ} 1:10 or {typ} 22', '{typ} 1:10 22'),
        ('{typ} 1:10 or {typ} 20:30', '{typ} 1:10 20:30'),
        ('{typ} 1-5 or {typ} 7', '{typ} 1-5 7'),
        ('{typ} 1-5 or {typ} 7:10 or {typ} 12', '{typ} 1-5 7:10 12'),
        ('{typ} 1 or {typ} 3 or {typ} 5:10', '{typ} 1 3 5:10'),
    ])
    def test_range_selections(self, seltype, ref, sel, universe):
        self._check_sels(ref.format(typ=seltype), sel.format(typ=seltype),
                         universe)


class TestICodeSelection(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(PDB_icodes)

    def test_select_icode(self, u):
        ag = u.select_atoms('resid 163A')

        assert len(ag) == 7
        assert_equal(ag.ids, np.arange(7) + 1230)

    def test_select_resid_implicit_icode(self, u):
        ag = u.select_atoms('resid 163')

        assert len(ag) == 6
        assert_equal(ag.ids, np.arange(6) + 1224)

    def test_select_icode_range_1(self, u):
        # testing range within a single resid integer value
        u = u
        ag = u.select_atoms('resid 163B-163D')

        # do it manually without selection language...
        ref = u.residues[u.residues.resids == 163]
        ref = ref[(ref.icodes >= 'B') & (ref.icodes <= 'D')]
        ref = ref.atoms

        assert_equal(ag.ids, ref.ids)

        assert len(ag) == 19
        assert_equal(ag.ids, np.arange(19) + 1237)

    def test_select_icode_range_2(self, u):
        u = u

        ag = u.select_atoms('resid 163B-165')

        resids = u.residues.resids

        start = u.residues[resids == 163]
        start = start[start.icodes >= 'B']

        mid = u.residues[resids == 164]

        end = u.residues[resids == 165]
        end = end[end.icodes == '']

        ref = start.atoms + mid.atoms + end.atoms

        assert_equal(ag.ids, ref.ids)

    def test_select_icode_range_3(self, u):
        # same as #2 but with no "middle" icodes
        u = u

        ag = u.select_atoms('resid 163B-164')

        resids = u.residues.resids

        start = u.residues[resids == 163]
        start = start[start.icodes >= 'B']

        end = u.residues[resids == 164]
        end = end[end.icodes == '']

        ref = start.atoms + end.atoms

        assert_equal(ag.ids, ref.ids)

    def test_select_icode_range_4(self, u):
        u = u

        ag = u.select_atoms('resid 160-163G')

        resids = u.residues.resids

        start = u.residues[resids == 160]
        start = start[start.icodes >= '']

        mid = u.residues[(resids == 161) | (resids == 162)]

        end = u.residues[resids == 163]
        end = end[end.icodes <= 'G']

        ref = start.atoms + mid.atoms + end.atoms

        assert_equal(ag.ids, ref.ids)

    def test_select_icode_range_5(self, u):
        # same as #4 but with no "middle" icodes in range
        u = u

        ag = u.select_atoms('resid 162-163G')

        resids = u.residues.resids

        start = u.residues[resids == 162]
        start = start[start.icodes >= '']

        end = u.residues[resids == 163]
        end = end[end.icodes <= 'G']

        ref = start.atoms + end.atoms

        assert_equal(ag.ids, ref.ids)

    def test_missing_icodes_VE(self, u):
        # trying a selection with icodes in a Universe without raises VA
        u = make_Universe(('resids',))
        with pytest.raises(ValueError):
            u.select_atoms('resid 10A')

    def test_missing_icodes_range_VE(self, u):
        u = make_Universe(('resids',))
        with pytest.raises(ValueError):
            u.select_atoms('resid 10A-12')


@pytest.fixture
def u_pdb_icodes():
    return mda.Universe(PDB_icodes)


@pytest.mark.parametrize(
    "selection, n_atoms",
    [
        # Selection using resindices
        # For PDBs:
        # residues with different insertion codes have different resindices
        ("same residue as ", 11),
        # Selection using resids
        # Residues with different insertion codes have the same resid
        # See Issues #2308 for a discussion
        ("same resid as", 72),
        # Selection using resindices
        # For PDBs: 
        # residues with different insertion codes have different resindices
        ("byres", 11)
    ]
)
def test_similarity_selection_icodes(u_pdb_icodes, selection, n_atoms):

    # Select residues 162 and 163A
    sel = u_pdb_icodes.select_atoms(selection + "(around 2.0 resid 163)")

    assert len(sel.atoms) == n_atoms

@pytest.mark.parametrize('selection', [
    'all', 'protein', 'backbone', 'nucleic', 'nucleicbackbone',
    'name O', 'name N*', 'resname stuff', 'resname ALA', 'type O',
    'index 0', 'index 1', 'bynum 1-10',
    'segid SYSTEM', 'resid 163', 'resid 1-10', 'resnum 2',
    'around 10 resid 1', 'point 0 0 0 10', 'sphzone 10 resid 1',
    'sphlayer 0 10 index 1', 'cyzone 15 4 -8 index 0',
    'cylayer 5 10 10 -8 index 1', 'prop abs z <= 100',
    'byres index 0', 'same resid as index 0',
])
def test_selections_on_empty_group(u_pdb_icodes, selection):
    ag = u_pdb_icodes.atoms[[]].select_atoms(selection)
    assert len(ag) == 0

def test_empty_yet_global(u_pdb_icodes):
    # slight exception to above test, an empty AG can return something if 'global' used
    ag = u_pdb_icodes.atoms[[]].select_atoms('global name O')

    assert len(ag) == 185  # len(u_pdb_icodes.select_atoms('name O'))

def test_arbitrary_atom_group_raises_error():
    u = make_Universe(trajectory=True)
    with pytest.raises(TypeError):
        u.select_atoms('around 2.0 group this', this=u.atoms[0])


def test_empty_sel():
    u = make_Universe(trajectory=True)
    with pytest.warns(UserWarning):
        ag = u.atoms.select_atoms("")
    assert_equal(len(ag), 0)
    assert isinstance(ag, mda.AtomGroup)


def test_record_type_sel():
    u = mda.Universe(PDB_HOLE)

    assert len(u.select_atoms('record_type ATOM')) == 264
    assert len(u.select_atoms('not record_type HETATM')) == 264
    assert len(u.select_atoms('record_type HETATM')) == 8

    assert len(u.select_atoms('name CA and not record_type HETATM')) == 30
    assert len(u.select_atoms('name CA and record_type HETATM')) == 2


def test_element_sel():
    # test auto-topattr addition of string
    u = mda.Universe(PDB_elements)
    assert len(u.select_atoms("element H")) == 8
    assert len(u.select_atoms("same element as index 12")) == 8


def test_chain_sel():
    u = mda.Universe(PDB_elements)
    assert len(u.select_atoms("chainID A")) == len(u.atoms)


@pytest.fixture()
def u_fake_masses():
    u = mda.Universe(TPR)
    u.atoms[-10:].masses = - (np.arange(10) + 1.001)
    u.atoms[:5].masses = 0.1 * 3  # 0.30000000000000004
    u.atoms[5:10].masses = 0.30000000000000001
    return u


@pytest.mark.parametrize("selstr,n_atoms, selkwargs", [
    ("mass 0.8 to 1.2", 23844, {}),
    ("mass 8e-1 to 1200e-3", 23844, {}),
    ("mass -5--3", 2, {}),  # select -5 to -3
    ("mass -3 : -5", 0, {}),  # wrong way around
    # regex; spacing, delimiters, and negatives
    ("mass -5 --3", 2, {}),
    ("mass -5- -3", 2, {}),
    ("mass -5 - -3", 2, {}),
    ("mass -10:3", 34945, {}),
    ("mass -10 :3", 34945, {}),
    ("mass -10: 3", 34945, {}),
    ("mass -10 : 3", 34945, {}),
    ("mass -10 -3", 0, {}),  # separate selections, not range
    # float equality
    ("mass 0.3", 5, {"rtol": 0, "atol": 0}),
    ("mass 0.3", 5, {"rtol": 1e-22, "atol": 1e-22}),
    # 0.30000000000000001 == 0.3
    ("mass 0.3 - 0.30000000000000004", 10, {}),
    ("mass 0.30000000000000004", 5, {"rtol": 0, "atol": 0}),
    ("mass 0.3 0.30000000000000001", 5, {"rtol": 0, "atol": 0}),
    # float near-equality
    ("mass 0.3", 10, {}),
    ("mass 0.30000000000000004", 10, {}),
    ("mass 0.3 0.30000000000000001", 10, {}),
    # prop thingy
    ("prop mass == 0.3", 10, {}),
    ("prop mass == 0.30000000000000004", 10, {}),
    ("prop mass == 0.30000000000000004", 5, {"rtol": 0, "atol": 0}),
])
def test_mass_sel(u_fake_masses, selstr, n_atoms, selkwargs):
    # test auto-topattr addition of float (FloatRangeSelection)
    ag = u_fake_masses.select_atoms(selstr, **selkwargs)
    assert len(ag) == n_atoms

def test_mass_sel_warning(u_fake_masses):
    warn_msg = (r"Using float equality .* is not recommended .* "
                r"we recommend using a range .*"
                r"'mass -0.6 to 1.4'.*"
                r"use the `atol` and `rtol` keywords")
    with pytest.warns(SelectionWarning, match=warn_msg):
        u_fake_masses.select_atoms("mass 0.4")


@pytest.mark.parametrize("selstr,n_res", [
    ("resnum -10 to 3", 13),
    ("resnum -5--3", 3),  # select -5 to -3
    ("resnum -3 : -5", 0),  # wrong way around
])
def test_int_sel(selstr, n_res):
    # test auto-topattr addition of int (IntRangeSelection)
    u = mda.Universe(TPR)
    u.residues[-10:].resnums = - (np.arange(10) + 1)
    ag = u.select_atoms(selstr).residues
    assert len(ag) == n_res


def test_negative_resid():
    # this is its own separate test because ResidSelection
    # has special matching for icodes
    text = """\
    ATOM      1  N   ASP A  -1      19.426  19.251   2.191  1.00 59.85           N
    ATOM      2  CA  ASP A  -1      20.185  18.441   1.255  1.00 54.54           C
    ATOM      3  C   ASP A  -1      21.660  18.901   1.297  1.00 40.12           C
    ATOM      4  O   ASP A  -1      21.958  19.978   1.829  1.00 35.85           O"""
    u = mda.Universe(StringIO(textwrap.dedent(text)), format="PDB")
    ag = u.select_atoms("resid -1")
    assert len(ag) == 4


@pytest.mark.parametrize("selstr, n_atoms", [
    ("aromaticity", 5),
    ("aromaticity true", 5),
    ("not aromaticity", 15),
    ("aromaticity False", 15),
])
def test_bool_sel(selstr, n_atoms):
    pytest.importorskip("rdkit.Chem")
    u = MDAnalysis.Universe.from_smiles("Nc1cc(C[C@H]([O-])C=O)c[nH]1")
    assert len(u.select_atoms(selstr)) == n_atoms


def test_bool_sel_error():
    pytest.importorskip("rdkit.Chem")
    u = MDAnalysis.Universe.from_smiles("Nc1cc(C[C@H]([O-])C=O)c[nH]1")
    with pytest.raises(SelectionError, match="'fragrant' is an invalid value"):
        u.select_atoms("aromaticity fragrant")


def test_error_selection_for_strange_dtype():
    with pytest.raises(ValueError, match="No base class defined for dtype"):
        MDAnalysis.core.selection.gen_selection_class("star", "stars",
                                                      dict, "atom")


@pytest.mark.parametrize("sel, ix", [
    ("name N", [5, 335, 451]),
    ("resname GLU", [5, 6, 7, 8, 335, 451]),
])
def test_default_selection_on_ordered_unique_group(u_pdb_icodes, sel, ix):
    # manually ordered unique atomgroup => sorted by index
    base_ag = u_pdb_icodes.atoms[[335, 5, 451, 8, 7, 6]]
    ag = base_ag.select_atoms(sel)
    assert_equal(ag.ix, ix)


@pytest.mark.parametrize("sel, sort, ix", [
    ("name N", True, [5, 335, 451]),
    ("name N", False, [335, 5, 451]),
    ("resname GLU", True, [5, 6, 7, 8, 335, 451]),
    ("resname GLU", False, [335, 5, 451, 8, 7, 6]),
])
def test_unique_selection_on_ordered_unique_group(u_pdb_icodes, sel, sort, ix):
    # manually ordered unique atomgroup
    base_ag = u_pdb_icodes.atoms[[335, 5, 451, 8, 7, 6]]
    ag = base_ag.select_atoms(sel, sorted=sort)
    assert_equal(ag.ix, ix)


@pytest.mark.parametrize("sel, sort, ix", [
    ("name N", True, [5, 335, 451]),
    ("name N", False, [335, 5, 451]),
    ("resname GLU", True, [5, 6, 7, 8, 335, 451]),
    ("resname GLU", False, [335, 5, 451, 8, 7, 6]),
])
def test_unique_selection_on_ordered_group(u_pdb_icodes, sel, sort, ix):
    # manually ordered duplicate atomgroup
    base_ag = u_pdb_icodes.atoms[[335, 5, 451, 8, 5, 5, 7, 6, 451]]
    ag = base_ag.select_atoms(sel, sorted=sort)
    assert_equal(ag.ix, ix)


@pytest.mark.parametrize('smi,chirality', [
    ('C[C@@H](C(=O)O)N', 'S'),
    ('C[C@H](C(=O)O)N', 'R'),
])
def test_chirality(smi, chirality):
    Chem = pytest.importorskip('rdkit.Chem', reason='requires rdkit')

    m = Chem.MolFromSmiles(smi)
    u = mda.Universe(m)

    assert hasattr(u.atoms, 'chiralities')

    assert u.atoms[0].chirality == ''
    assert u.atoms[1].chirality == chirality

    assert_equal(u.atoms[:3].chiralities, np.array(['', chirality, ''], dtype='U1'))


@pytest.mark.parametrize('sel,size', [
    ('R', 1), ('S', 1), ('R S', 2), ('S R', 2),
])
def test_chirality_selection(sel, size):
    # 2 centers, one R one S
    Chem = pytest.importorskip('rdkit.Chem', reason='requires rdkit')

    m = Chem.MolFromSmiles('CC[C@H](C)[C@H](C(=O)O)N')
    u = mda.Universe(m)

    ag = u.select_atoms('chirality {}'.format(sel))

    assert len(ag) == size



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import MDAnalysis as mda
import MDAnalysis.analysis.psa as PSA

from numpy.testing import (assert_array_less,
                           assert_almost_equal, assert_equal)
import numpy as np
import scipy
import scipy.spatial
import matplotlib

from MDAnalysisTests.datafiles import PSF, DCD, DCD2
from MDAnalysis import NoDataError


class TestPSAnalysis(object):
    iu1 = np.triu_indices(3, k=1)

    @pytest.fixture()
    def psa(self, tmpdir):
        universe1 = mda.Universe(PSF, DCD)
        universe2 = mda.Universe(PSF, DCD2)
        universe_rev = mda.Universe(PSF, DCD)

        psa = PSA.PSAnalysis([universe1, universe2, universe_rev],
                             path_select='name CA',
                             targetdir=str(tmpdir))

        psa.generate_paths(align=True)
        psa.paths[-1] = psa.paths[-1][::-1, :, :]  # reverse third path
        return psa

    @pytest.fixture()
    def hausd_matrix(self, psa):
        psa.run(metric='hausdorff')
        return psa.get_pairwise_distances()

    @pytest.fixture()
    def hausd_dists(self, hausd_matrix):
        return hausd_matrix[self.iu1]

    @pytest.fixture()
    def frech_matrix(self, psa):
        psa.run(metric='discrete_frechet')
        return psa.get_pairwise_distances()

    @pytest.fixture()
    def frech_dists(self, frech_matrix):
        return frech_matrix[self.iu1]

    @pytest.fixture()
    def plot_data(self, psa, tmpdir):
        psa.run(metric='hausdorff')
        psa.run(metric='discrete_frechet')
        with tmpdir.as_cwd():
            results = psa.plot(filename="distmat.png")
        return results

    @pytest.fixture()
    def plot_annotated_heatmap(self, psa, tmpdir):
        pytest.importorskip('seaborn')
        psa.run(metric='hausdorff')
        with tmpdir.as_cwd():
            results = psa.plot_annotated_heatmap(filename="annotated.png")
        return results

    @pytest.fixture()
    def plot_nearest_neighbors(self, psa, tmpdir):
        pytest.importorskip('seaborn')
        psa.run(metric='hausdorff')
        psa.run_pairs_analysis(neighbors=True)
        with tmpdir.as_cwd():
            results = psa.plot_nearest_neighbors(filename="nn.png")
        return results

    @pytest.fixture()
    def hausd_pairs_dists(self, psa):
        psa.run_pairs_analysis(neighbors=True, hausdorff_pairs=True)
        hausd_pairs_matrix = psa.get_pairwise_distances()

        return hausd_pairs_matrix[self.iu1]

    def test_hausdorff_bound(self, hausd_dists, frech_dists):
        """Test whether Frechet distances are smaller than corresponding
        Hausdorff distances"""
        err_msg = ("Some Frechet distances are smaller than corresponding "
                   "Hausdorff distances")
        assert_array_less(hausd_dists, frech_dists, err_msg)

    def test_explicit_metric(self, psa, hausd_dists):
        """Test whether explicitly specifying Hausdorff metric gives same result
        as specifying Hausdorff metric with string name"""
        psa.run(metric=PSA.hausdorff)
        hausd_matrix_explicit = psa.get_pairwise_distances()
        hausd_explicit_dists = hausd_matrix_explicit[self.iu1]

        err_msg = ("Specifying Python function for Hausdoff gives different "
                   "distances than specifying Hausdorff with string name")
        assert_equal(hausd_dists, hausd_explicit_dists, err_msg)

    def test_reversal_hausdorff(self, hausd_matrix):
        """Test whether Hausdorff distances are invariant to path reversal"""
        err_msg = "Hausdorff distances changed after path reversal"
        assert_almost_equal(hausd_matrix[1,2],
                            hausd_matrix[0,1],
                            decimal=3, err_msg=err_msg)

    def test_reversal_frechet(self, frech_matrix):
        """Test whether Frechet distances are same/larger after path reversal"""
        err_msg = "Frechet distances did not increase after path reversal"
        assert frech_matrix[1,2] >= frech_matrix[0,1], err_msg

    def test_get_num_paths(self, psa):
        assert psa.get_num_paths() == 3

    def test_get_paths(self, psa):
        paths = psa.get_paths()
        assert len(paths) == 3
        assert isinstance(paths, list)

    def test_psa_pairs_ValueError(self, psa):
        with pytest.raises(ValueError):
            psa.psa_pairs

    def test_psa_pairs(self, psa):
        psa.run_pairs_analysis()
        assert len(psa.psa_pairs) == 3

    def test_hausdorff_pairs_ValueError(self, psa):
        with pytest.raises(ValueError):
            psa.hausdorff_pairs

    def test_hausdorff_pairs(self, psa):
        psa.run_pairs_analysis(hausdorff_pairs=True)
        assert len(psa.hausdorff_pairs) == 3

    def test_nearest_neighbors_ValueError(self, psa):
        with pytest.raises(ValueError):
            psa.nearest_neighbors

    def test_nearest_neighbors(self, psa):
        psa.run_pairs_analysis(neighbors=True)
        assert len(psa.nearest_neighbors) == 3

    @pytest.mark.parametrize('stored', [True, False])
    def test_load(self, stored, tmpdir):
        """Test that the automatically saved files can be loaded"""
        # To allow for testing the store keyword, ignore fixture
        universe1 = mda.Universe(PSF, DCD)
        universe2 = mda.Universe(PSF, DCD2)
        universe_rev = mda.Universe(PSF, DCD)

        psa = PSA.PSAnalysis([universe1, universe2, universe_rev],
                              path_select='name CA',
                              targetdir=str(tmpdir))

        psa2 = PSA.PSAnalysis([universe1, universe2, universe_rev],
                              path_select='name CA',
                              targetdir=str(tmpdir))

        psa.generate_paths(align=True, store=stored)

        # Make copies to the existing data
        # Note: path names are set after save_paths has been called
        expected_paths = [p.copy() for p in psa.paths]

        if not stored:
            psa.save_paths()

        expected_path_names = psa.path_names[:]

        # Load data in the empty PSAnalysis object
        psa2.load()

        assert psa2.path_names == expected_path_names
        assert len(psa2.paths) == len(expected_paths)

        for ipath, (observed, expected) in enumerate(zip(psa2.paths,
                                                         expected_paths)):
            assert_almost_equal(observed, expected, decimal=6,
                                err_msg=("loaded path {} does not agree with "
                                         "input").format(ipath))

    def test_load_nofile(self, psa):
        """Test case where save_paths hasn't been called before load"""
        match_exp = "Fitted trajectories cannot be loaded"
        with pytest.raises(NoDataError, match=match_exp):
            psa.load()

    def test_save_nopaths(self, tmpdir):
        """Test case were save_paths is called without calcualted paths"""
        match_exp = "Paths have not been calculated yet"
        with pytest.raises(NoDataError, match=match_exp):
            universe1 = mda.Universe(PSF, DCD)
            universe2 = mda.Universe(PSF, DCD2)
            universe_rev = mda.Universe(PSF, DCD)

            psa = PSA.PSAnalysis([universe1, universe2, universe_rev],
                                 path_select='name CA',
                                 targetdir=str(tmpdir))
            psa.save_paths()

    def test_dendrogram_produced(self, plot_data):
        """Test whether Dendrogram dictionary object was produced"""
        err_msg = "Dendrogram dictionary object was not produced"
        assert isinstance(plot_data[1], dict), err_msg

    def test_dendrogram_produced_annotated(self, plot_annotated_heatmap):
        """Test whether Dendrogram dictionary object was produced"""
        err_msg = "Dendrogram dictionary object was not produced"
        assert isinstance(plot_annotated_heatmap[1], dict), err_msg

    def test_plot_nearest_neighbors(self, plot_nearest_neighbors):
        assert isinstance(plot_nearest_neighbors, matplotlib.axes.Axes)

    def test_dist_mat_to_vec_i_less_j(self):
        """Test the index of corresponding distance vector is correct if i < j"""
        err_msg = "dist_mat_to_vec function returning wrong values"
        assert_equal(PSA.dist_mat_to_vec(5, 3, 4), 9, err_msg)

    def test_dist_mat_to_vec_i_greater_j(self):
        """Test the index of corresponding distance vector is correct if i > j"""
        err_msg = "dist_mat_to_vec function returning wrong values"
        assert_equal(PSA.dist_mat_to_vec(5, 4, 3), 9, err_msg)

    def test_dist_mat_to_vec_input_numpy_integer_32(self):
        """Test whether inputs are supported as numpy integers rather than normal Integers"""
        err_msg = "dist_mat_to_vec function returning wrong values"
        assert_equal(
            PSA.dist_mat_to_vec(np.int32(5), np.int32(3), np.int32(4)), np.int32(9),
            err_msg)

    def test_dist_mat_to_vec_input_numpy_integer_16(self):
        """Test whether inputs are supported as numpy integers rather than normal Integers"""
        err_msg = "dist_mat_to_vec function returning wrong values"
        assert_equal(
            PSA.dist_mat_to_vec(np.int16(5), np.int16(3), np.int16(4)), np.int16(9),
            err_msg)

    def test_hausdorff_pairs_distances(self, hausd_dists, hausd_pairs_dists):
        """Test whether Hausdorff pairs analysis distances are
        identical to those from standard Hausdorff metric"""
        err_msg = ("Some Hausdorff distances from pairs analysis vary "
                   "significantly from usual Hausdorff calculation")
        assert_almost_equal(hausd_dists, hausd_pairs_dists,
                            decimal=6, err_msg=err_msg)

    def test_distances_from_hausdorff_pairs_frames(self, psa):
        """Test whether actual distances between frames of Hausdorff
        pairs of a path give the expected Hausdorff distance"""
        psa.run_pairs_analysis(neighbors=True, hausdorff_pairs=True)
        hausd_pairs = psa.hausdorff_pairs
        npairs = int(psa.npaths * (psa.npaths - 1) / 2)
        hausd_pairs_dists2 = np.array([hausd_pairs[i]['distance']
                                       for i in range(npairs)])


        err_msg = ("A Hausdorff pair analysis distance when accessed "
                   "by frame varies from expected Hausdorff distance")
        dists = np.zeros((psa.npaths, psa.npaths))
        for i in range(0, psa.npaths-1):
            for j in range(i+1, psa.npaths):
                pairidx = PSA.dist_mat_to_vec(psa.npaths, i, j)
                p, q = hausd_pairs[pairidx]['frames']
                dists[i,j] = (PSA.sqnorm(psa.paths[i][p,:,:] -
                                         psa.paths[j][q,:,:]) /
                                         psa.natoms)**0.5
        assert_almost_equal(hausd_pairs_dists2,
                            dists[self.iu1],
                            decimal=6, err_msg=err_msg)

class TestPSAExceptions(object):
    '''Tests for exceptions that should be raised
    or caught by code in the psa module.'''

    def test_get_path_metric_func_bad_key(self):
        '''Test that KeyError is caught by
        get_path_metric_func().'''
        with pytest.raises(KeyError):
            PSA.get_path_metric_func('123456')

    def test_get_coord_axes_bad_dims(self):
        """Test that ValueError is raised when
        numpy array with incorrect dimensions
        is fed to get_coord_axes()."""
        with pytest.raises(ValueError):
            PSA.get_coord_axes(np.zeros((5,5,5,5)))

    @pytest.mark.parametrize('N, i, j', (
        (5, 6, 4),
        (5, 4, 6),
        (5, 6, 7),
        (5, -1, 2),
        (5, 1, -2),
        (1, 0, 0)

    ))
    def test_dist_mat_to_vec_func_out_of_bounds(self, N, i, j):
        """Test that ValueError is raised when i or j or both are
        out of bounds of N"""

        # Check if i is out of bounds of N
        with pytest.raises(ValueError):
            PSA.dist_mat_to_vec(N, i, j)

    @pytest.mark.parametrize('N, i, j', (
        (5, 4, 4),
        (4, 6, 4)
    ))
    def test_dist_mat_to_vec_func_i_equals_j(self, N, i, j):
        """Test that ValueError is raised when i == j or i,j == N"""

        with pytest.raises(ValueError):
            PSA.dist_mat_to_vec(N, i, j)

    def test_dist_mat_to_vec_func_bad_integers(self):
        """Test that ValueError is raised when i or j are
        not Integers"""

        with pytest.raises(ValueError) as err:
            PSA.dist_mat_to_vec(5, '6', '7')
        assert 'all must be of type int' in str(err.value)

        with pytest.raises(ValueError):
            PSA.dist_mat_to_vec(5, float(6), 7)


class _BaseHausdorffDistance(object):
    '''Base Class setup and unit tests
    for various Hausdorff distance
    calculation properties.'''

    @pytest.fixture()
    def random_angles(self):
        return np.random.random((100,)) * np.pi * 2

    @staticmethod
    @pytest.fixture()
    def path_1(random_angles):
        random_columns = np.column_stack((random_angles,
                                          random_angles,
                                          np.zeros((100,))))
        random_columns[..., 0] = np.cos(random_columns[..., 0])
        random_columns[..., 1] = np.sin(random_columns[..., 1])

        return random_columns

    @staticmethod
    @pytest.fixture()
    def path_2(random_angles):
        random_columns_2 = np.column_stack((random_angles,
                                            random_angles,
                                            np.zeros((100,))))
        random_columns_2[1:, 0] = np.cos(random_columns_2[1:, 0]) * 2.0
        random_columns_2[1:, 1] = np.sin(random_columns_2[1:, 1]) * 2.0
        # move one point farther out so we don't have two perfect circles
        random_columns_2[0, 0] = np.cos(random_columns_2[0, 0]) * 3.3
        random_columns_2[0, 1] = np.sin(random_columns_2[0, 1]) * 3.3
        return random_columns_2

    def test_symmetry(self, path_1, path_2, h):
        '''Ensure that the undirected (symmetric)
        Hausdorff distance is actually symmetric
        for a given Hausdorff metric, h.'''
        forward = h(path_1, path_2)
        reverse = h(path_2, path_1)
        # lower precision on 32bit
        assert_almost_equal(forward, reverse, decimal=15)

    def test_hausdorff_value(self, path_1, path_2, h, expected):
        '''Test that the undirected Hausdorff
        distance matches expected value for
        the simple case here.'''
        actual = h(path_1, path_2)
        # unless I pin down the random generator
        # seems unstable to use decimal > 2
        assert_almost_equal(actual, expected, decimal=2)


class TestHausdorffSymmetric(_BaseHausdorffDistance):
    '''Tests for conventional and symmetric (undirected)
    Hausdorff distance between point sets in 3D.'''

    # expected = 2.3

    @pytest.fixture()
    def expected(self):
        return 2.3

    @pytest.fixture()
    def h(self):
        return PSA.hausdorff


class TestWeightedAvgHausdorffSymmetric(_BaseHausdorffDistance):
    '''Tests for weighted average and symmetric (undirected)
    Hausdorff distance between point sets in 3D.'''

    @pytest.fixture()
    def expected(self, path_1, path_2):
        distance_matrix = scipy.spatial.distance.cdist(path_1, path_2)
        return (np.mean(np.amin(distance_matrix, axis=0)) +
                np.mean(np.amin(distance_matrix, axis=1))) / 2.0

    # params instead of 2 (I think it sends self as a parameter too.)
    @pytest.fixture()
    def h(self):
        return PSA.hausdorff_wavg

    def test_asymmetric_weight(self, path_1, path_2, h):
        '''Test for WAvg Hausdorff to ensure that increasing N points in one
        of the paths does NOT increase the weight of its contributions.'''
        inflated_path_1 = np.concatenate((path_1, path_1))
        inflated_path_2 = np.concatenate((path_2, path_2))
        d_inner_inflation = h(inflated_path_1, path_2)
        d_outer_inflation = h(path_1, inflated_path_2)
        assert_almost_equal(d_inner_inflation, d_outer_inflation)


class TestAvgHausdorffSymmetric(_BaseHausdorffDistance):
    '''Tests for unweighted average and symmetric (undirected)
    Hausdorff distance between point sets in 3D.'''

    @pytest.fixture()
    def expected(self, path_1, path_2):
        distance_matrix = scipy.spatial.distance.cdist(path_1, path_2)
        return np.mean(np.append(np.amin(distance_matrix, axis=0),
                                 np.amin(distance_matrix, axis=1)))

    # params instead of 2 (I think it sends self as a parameter too.)
    @pytest.fixture()
    def h(self):
        return PSA.hausdorff_avg

    def test_asymmetric_weight(self, path_1, path_2, h):
        '''Test to ensure that increasing N points in one of the paths
        increases the weight of its contributions.'''
        inflated_path_1 = np.concatenate((path_1, path_1))
        inflated_path_2 = np.concatenate((path_2, path_2))
        d_inner_inflation = h(inflated_path_1, path_2)
        d_outer_inflation = h(path_1, inflated_path_2)

        assert_array_less(d_inner_inflation, d_outer_inflation)


class DiscreteFrechetDistance(object):
    @staticmethod
    @pytest.fixture()
    def random_angles():
        return np.random.random((100,)) * np.pi * 2

    @staticmethod
    @pytest.fixture()
    def path_1(random_angles):
        random_columns = np.column_stack((random_angles, random_angles,
                                          np.zeros((100,))))
        random_columns[..., 0] = np.cos(random_columns[..., 0])
        random_columns[..., 1] = np.sin(random_columns[..., 1])
        return random_columns

    @staticmethod
    @pytest.fixture()
    def path_2(random_angles):
        random_columns_2 = np.column_stack((random_angles, random_angles,
                                            np.zeros((100,))))
        random_columns_2[..., 0] = np.cos(random_columns_2[..., 0]) * 5.5
        random_columns_2[..., 1] = np.sin(random_columns_2[..., 1]) * 5.5
        return random_columns_2

    def test_discrete_Frechet_concentric_circles(self, path_1, path_2):
        # test for the simple case of the discrete Frechet distance
        # between concentric circular paths, which for a sufficiently
        # high random discrete point density around each circle
        # should be the absolute difference between their respective
        # radii

        expected = 4.5
        actual = PSA.discrete_frechet(path_1, path_2)
        assert_almost_equal(actual, expected)
